# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
from functools import partial, wraps

from pydiverse.transform._internal.errors import SubqueryError
from pydiverse.transform._internal.tree import verbs
from pydiverse.transform._internal.tree.col_expr import Col


class Pipeable:
    def __init__(self, f=None, calls=None):
        if f is not None:
            if calls is not None:
                raise ValueError
            self.calls = [f]
        else:
            self.calls = calls

    def __rshift__(self, other) -> "Pipeable":
        """
        The pipe operator for chaining verbs.
        """

        if isinstance(other, Pipeable):
            return Pipeable(calls=self.calls + other.calls)
        elif callable(other):
            return Pipeable(calls=self.calls + [other])

        raise RuntimeError

    def __call__(self, arg):
        for c in self.calls:
            res = c(arg)
            if isinstance(res, Pipeable):
                res = res(arg)
            arg = res
        return arg


class inverse_partial(partial):
    """
    Just like partial, but the arguments get applied to the back instead of the front.
    This means that a function `def x(a, b, c)` decorated with `@inverse_partial(1, 2)`
    that gets called with `x(0)` is equivalent to calling `x(0, 1, 2)` on the non
    decorated function.
    """

    def __call__(self, /, *args, **keywords):
        keywords = {**self.keywords, **keywords}
        return self.func(*args, *self.args, **keywords)


def verb(fn):
    """
    Decorator for creating verbs.

    A verb is simply a function decorated with this decorator that takes a
    pydiverse.transform Table as the first argument. `@verb` enables usage of
    the function with the pipe `>>` syntax.

    Examples
    --------
    >>> @verb
    ... def strip_all_strings(tbl: pdt.Table) -> pdt.Table:
    ...     return tbl >> mutate(
    ...         **{c.name: c.str.strip() for c in tbl if c.dtype() == pdt.String()}
    ...     )
    >>> t = pdt.Table(
    ...     {
    ...         "a": ["  abcd 5 ", "212"],
    ...         "b": [" 917. __ ", " 2 "],
    ...         "c": [1, 2],
    ...     },
    ...     name="t",
    ... )
    >>> t >> strip_all_strings() >> show()
    Table `t` (backend: polars)
    shape: (2, 3)
    ┌─────┬────────┬─────────┐
    │ c   ┆ a      ┆ b       │
    │ --- ┆ ---    ┆ ---     │
    │ i64 ┆ str    ┆ str     │
    ╞═════╪════════╪═════════╡
    │ 1   ┆ abcd 5 ┆ 917. __ │
    │ 2   ┆ 212    ┆ 2       │
    └─────┴────────┴─────────┘

    Note
    ----
    To make the code completion of your IDE not show the table as the first argument,
    simply add an `@overload` before the verb definition:
    >>> @overload
    ... def strip_all_strings() -> pdt.Table: ...
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        return Pipeable(inverse_partial(fn, *args, **kwargs))

    return wrapper


def check_subquery(new_tbl, child_tbl, *, is_right: bool = False):
    if (reason := child_tbl._cache.requires_subquery(new_tbl._ast)) is not None:
        # Search among descendants of the current node for an `Alias` that can be used
        # to create a subquery. If we hit a `Join` or `SubqueryMarker`, we stop.
        chain: list[verbs.Verb] = [new_tbl._ast]
        for nd in child_tbl._ast.iter_subtree_preorder():
            if isinstance(nd, verbs.Alias):
                new_chain = [copy.copy(c) for c in chain]
                new_chain.append(verbs.SubqueryMarker(nd))

                # rebuild the part of the AST leading to the `Alias`
                for i in range(1, len(new_chain) - 1):
                    new_chain[i].child = new_chain[i + 1]
                if is_right:
                    assert isinstance(new_tbl._ast, verbs.Join)
                    new_chain[0].right = new_chain[1]
                else:
                    new_chain[0].child = new_chain[1]

                from pydiverse.transform._internal.pipe.table import Table

                # See if we still need a subquery
                test_tbl = Table(new_chain[1])
                new_chain[0].map_col_nodes(
                    lambda expr: test_tbl._cache.cols[expr._uuid]  # noqa: B023
                    if isinstance(expr, Col) and expr._uuid in test_tbl._cache.cols  # noqa: B023
                    else expr
                )
                if test_tbl._cache.requires_subquery(new_chain[0]):
                    break

                modified_new_tbl = copy.copy(new_tbl)
                modified_new_tbl._ast = new_chain[0]
                return (modified_new_tbl, test_tbl)

            if isinstance(nd, verbs.SubqueryMarker | verbs.Join):
                break
            chain.append(nd)

        raise SubqueryError(
            f"Executing the `{new_tbl._ast.__class__.__name__.lower()}` verb "
            f"on the table `{new_tbl._ast.child.name}` requires a subquery, which "
            "is forbidden in transform by default.\n"
            f"reason for the subquery: {reason}\n"
            f"hint: Materialize the table `{new_tbl._ast.child.name}` before this "
            "verb. If you are sure you want to do a subquery, put an "
            "`>> alias()` before this verb. "
        )

    return (new_tbl, child_tbl)


# Checks for subqueries and updates the cache for all verbs except `join`. Since `join`
# is much more complex, we do these tasks manually in the verb.
def modify_ast(fn):
    @wraps(fn)
    def _fn(table, *args, **kwargs):
        new = fn(table, *args, **kwargs)
        assert new._ast != table._ast

        assert new._ast.child == table._ast
        new, child = check_subquery(new, table)
        new._cache = child._cache.update(new._ast)

        return new

    return _fn



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform.extended import *


def test_preview_print(df3, df4, df_strings):
    def data_part(p: str):
        return p.split("\n", 1)[1]

    assert data_part(str(df3[0])) == data_part(str(df3[1]))
    assert data_part(str(df_strings[0])) == data_part(str(df_strings[1]))

    long_tbl = [
        # after a join, the order is arbitrary, so we need to enforce a specific order
        left >> cross_join(right) >> arrange(left.col4, right.col4.nulls_last())
        for left, right in zip(df3, df4, strict=True)
    ]
    assert data_part(str(long_tbl[0])) == data_part(str(long_tbl[1]))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import logging

import pytest

from pydiverse.common.util.structlog import setup_logging

# Setup


supported_options = [
    "duckdb_parquet",
    "sqlite",
    "postgres",
    "mssql",
    "ibm_db2",
]


def pytest_addoption(parser):
    for opt in supported_options:
        parser.addoption(
            "--" + opt,
            action="store_true",
            default=False,
            help=f"run test that require {opt}",
        )


def pytest_collection_modifyitems(config: pytest.Config, items):
    for opt in supported_options:
        if not config.getoption("--" + opt):
            skip = pytest.mark.skip(reason=f"{opt} not selected")
            for item in items:
                if opt in item.keywords or any(
                    kw.startswith(f"{opt}-") or kw.endswith(f"-{opt}") or f"-{opt}-" in kw for kw in item.keywords
                ):
                    item.add_marker(skip)


setup_logging(log_level=logging.INFO)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import sqlite3

import pytest

import pydiverse.transform as pdt
from pydiverse.transform.errors import FunctionTypeError
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "full",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_join(df1, df2, how):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> join(u, t.col1 == u.col1, how=how),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> join(u, on="col1", how=how),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: u
        >> join(
            s := t >> left_join(u, on=t.col1 == u.col2) >> alias("b"),
            on=s.col1 == u.col1,
            how=how,
        ),
    )

    assert_result_equal(
        df1,
        lambda t: t
        >> left_join(
            r := t >> left_join(s := t >> alias("s"), on=t.col1 == s.col1) >> alias("r"),
            on=t.col1 == r.col1,
        ),
    )

    assert_result_equal(
        df1,
        lambda t: t
        >> left_join(s := t >> alias("s"), on=t.col1 == s.col1)
        >> left_join(r := t >> alias("r"), on=t.col1 == r.col1),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t
        >> join(u, (t.col1 == u.col1) & (t.col1 == u.col2), how=how)
        >> mutate(l=t.col2.str.len())
        >> left_join(v := u >> alias("v"), v.col2 == C.l)
        >> mutate(k=v.col1 + C.l + u.col2)
        >> full_join(w := t >> alias("w"), (t.col1 == w.col1) & (C.k == w.col1)),
        check_row_order=False,
    )


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "full",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_join_and_select(df1, df2, how):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> select() >> join(u, t.col1 == u.col1, how=how),
        check_row_order=False,
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> join(u >> select(), (t.col1 == u.col1) & (u.col2 == t.col1), how=how),
        check_row_order=False,
    )


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "full",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_self_join(df3, how):
    # Self join without alias should raise an exception
    assert_result_equal(
        df3,
        lambda t: t >> join(t, t.col1 == t.col1, how=how),
        exception=ValueError,
    )

    def self_join_1(t):
        u = t >> alias("self_join")
        return t >> join(u, t.col1 == u.col1, how=how)

    assert_result_equal(df3, self_join_1, check_row_order=False)

    def self_join_2(t):
        u = t >> alias("self_join")
        return t >> join(u, (t.col1 == u.col1) & (u.col2 == t.col2), how=how)

    assert_result_equal(df3, self_join_2, check_row_order=False)

    def self_join_3(t):
        u = t >> alias("self_join")
        return t >> join(u, (t.col2 == u.col3), how=how)

    assert_result_equal(df3, self_join_3, check_row_order=False)


def test_ineq_join(df3, df4, df_strings):
    assert_result_equal((df3, df4), lambda t, s: t >> inner_join(s, t.col1 <= s.col1))

    assert_result_equal(
        (df4, df_strings),
        lambda s, t: s
        >> inner_join(
            tm := t >> mutate(u=t.col1.str.len()),
            s.col3 <= tm.u,
            suffix="_42",
        ),
    )

    assert_result_equal(
        (df3, df_strings),
        lambda s, t: s
        >> inner_join(
            t,
            (s.col1 - s.col2 <= t.c.str.len())
            & (s.col4 >= pdt.when(t.col1.str.starts_with("-")).then(100).otherwise(4)),
        ),
    )

    assert_result_equal(
        (df3, df_strings),
        lambda s, t: s
        >> left_join(
            t,
            (s.col1 - s.col2 <= t.c.str.len() * t.d.str.len())
            & (s.col4 >= pdt.when(t.col1.str.starts_with("-")).then(100).otherwise(7)),
        ),
    )

    assert_result_equal(
        (df3, df_strings),
        lambda s, t: s
        >> left_join(
            t,
            (s.col4 - s.col2 <= t.col1.str.len() * t.d.str.len())
            & (s.col4 >= pdt.when(t.col1.str.starts_with(" ")).then(10).otherwise(7)),
        ),
    )

    assert_result_equal((df3, df4), lambda s, t: s >> inner_join(t, on=["col1", s.col2 <= t.col2]))


def test_join_summarize(df3, df4):
    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3 >> group_by(t3.col2) >> summarize(j=t3.col4.sum()) >> alias() >> inner_join(t4, on="col2"),
    )

    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t4 >> left_join(t3 >> group_by(t3.col2) >> summarize(j=t3.col4.sum()) >> alias(), on="col2"),
    )

    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3
        >> summarize(y=t3.col1.max(), z=t3.col4.mean())
        >> alias()
        >> left_join(t4, on=C.y == t4.col4),
    )


def test_join_window(df3, df4):
    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3 >> mutate(y=t3.col1.dense_rank()) >> alias() >> inner_join(t4, on=C.y == t4.col1),
    )

    assert_result_equal(
        (df3, df4),
        lambda s, t: s
        >> mutate(y=s.col1.shift(1, arrange=s.col4.nulls_first()))
        >> alias()
        >> inner_join(t, on="col2"),
    )

    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3 >> inner_join(t4, on=t3.col1.dense_rank() == t4.col1),
        exception=FunctionTypeError,
    )


def test_join_where(df2, df3, df4):
    assert_result_equal(
        (df2, df3),
        lambda t2, t3: t2 >> left_join(t3 >> filter(t3.col4 >= 2), on="col1"),
    )

    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3 >> filter(t3.col4 != -1729) >> left_join(t4 >> filter(t4.col3 > 0), on=t3.col2 == t4.col2),
    )

    assert_result_equal(
        (df3, df4),
        lambda t3, t4: t3
        >> filter(t3.col1 % 2 == 0)
        >> alias(keep_col_refs=True)
        >> full_join(
            t4_filtered := t4 >> filter(t4.col1.is_not_null()) >> alias(),
            on=t3.col2 == t4_filtered.col2,
        ),
    )


def test_join_const_col(df3, df4):
    assert_result_equal(
        (df3, df4),
        lambda s, t: s >> left_join(t >> mutate(y=0) >> alias(), on=s.col1 == C.y),
    )

    assert_result_equal(
        (df3, df4),
        lambda s, t: s >> mutate(z=2) >> alias() >> full_join(t >> mutate(j=True) >> alias(), on="col2"),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=4, y=5)
        >> mutate(z=C.x + C.y)
        >> alias(keep_col_refs=True)
        >> full_join(t_ := t >> alias(), on=t.col1 == t_.col2),
    )


def test_cross_join(df2, df3):
    assert_result_equal((df2, df3), lambda s, t: s >> cross_join(t))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt

import polars as pl
import pytest
import structlog

import pydiverse.transform as pdt
from pydiverse.transform._internal.errors import ColumnNotFoundError, DataTypeError
from pydiverse.transform.extended import *
from tests.util import assert_equal

logger = structlog.get_logger(__name__)

df1 = pl.DataFrame(
    {
        "col1": [1, 2, 3, 4],
        "col2": ["a", "b", "c", "d"],
    }
)

df2 = pl.DataFrame(
    {
        "col1": [1, 2, 2, 4, 5, 6],
        "col2": [2, 2, 0, 0, 2, None],
        "col3": [0.0, 0.1, 0.2, 0.3, 0.01, 0.02],
    }
)

df3 = pl.DataFrame(
    {
        "col1": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
        "col3": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        "col5": list("abcdefghijkl"),
    }
)

df4 = pl.DataFrame(
    {
        "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, None, 1, 0, 0, 1, 1],
        "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [None, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "col5": list("abcdefghijkl") + [None],
    }
)

df_left = pl.DataFrame(
    {
        "a": [1, 2, 3, 4],
    }
)

df_right = pl.DataFrame(
    {
        "b": [0, 1, 2, 2],
        "c": [5, 6, 7, 8],
    }
)

df_dt = pl.DataFrame(
    {
        "dt1": [
            dt.datetime(1999, 10, 31, hour=15),
            dt.datetime(2013, 7, 14, second=13, minute=55),
        ],
        "dur1": [dt.timedelta(days=5), dt.timedelta(minutes=4, microseconds=197)],
        "dt2": [dt.datetime(1974, 12, 31), dt.datetime(2020, 2, 14, microsecond=43)],
        "d1": [dt.date(2002, 6, 28), dt.date(2003, 1, 1)],
    }
)


@pytest.fixture
def dtype_backend(request):
    return request.param


@pytest.fixture
def tbl1():
    return pdt.Table(df1, name="df1")


@pytest.fixture
def tbl2():
    return pdt.Table(df2, name="df2")


@pytest.fixture
def tbl3():
    return pdt.Table(df3, name="df3")


@pytest.fixture
def tbl4():
    return pdt.Table(df4, name="df4")


@pytest.fixture
def tbl_left():
    return pdt.Table(df_left, name="df_left")


@pytest.fixture
def tbl_right():
    return pdt.Table(df_right, name="df_right")


@pytest.fixture
def tbl_dt():
    return pdt.Table(df_dt)


class TestPolarsLazyImpl:
    def test_dtype(self, tbl1, tbl2):
        assert isinstance(tbl1.col1.dtype(), pdt.Int64)
        assert isinstance(tbl1.col2.dtype(), pdt.String)

        assert isinstance(tbl2.col1.dtype(), pdt.Int64)
        assert isinstance(tbl2.col2.dtype(), pdt.Int64)
        assert isinstance(tbl2.col3.dtype(), pdt.Float64)

        # test that column expression type errors are checked immediately
        with pytest.raises(DataTypeError):
            tbl1.col1 + tbl1.col2

        # here, transform should not be able to resolve the type and throw an error
        C.col1 + tbl1.col2

    def test_build_query(self, tbl1):
        assert (tbl1 >> build_query()) is None

    def test_export(self, tbl1):
        # TODO: test export to other backends
        assert_equal(tbl1, df1)

    def test_select(self, tbl1):
        assert_equal(tbl1 >> select(tbl1.col1), df1.select("col1"))
        assert_equal(tbl1 >> select(tbl1.col2), df1.select("col2"))
        assert_equal(tbl1 >> select(), df1.select())

    def test_mutate(self, tbl1):
        assert_equal(
            tbl1 >> select() >> mutate(col1=4) >> mutate(col1=C.col1 + tbl1.col1),
            df1.with_columns(col1=pl.col("col1") + 4).select("col1"),
        )

        assert_equal(
            tbl1 >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1": [1, 2, 3, 4],
                    "col2": ["a", "b", "c", "d"],
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        assert_equal(
            tbl1 >> select() >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        # # Check proper column referencing
        t = tbl1 >> mutate(col2=tbl1.col1, col1=tbl1.col2)
        assert_equal(
            t >> select() >> mutate(x=t.col1, y=t.col2),
            tbl1 >> select() >> mutate(x=tbl1.col2, y=tbl1.col1),
        )
        assert_equal(
            t >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
            tbl1 >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
        )

    def test_join(self, tbl_left, tbl_right):
        assert_equal(
            tbl_left >> join(tbl_right, tbl_left.a == tbl_right.b, "left") >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2, 3, 4], "b": [1, 2, 2, None, None]}),
            check_row_order=False,
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "inner", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2], "b": [1, 2, 2]}),
            check_row_order=False,
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "full", suffix="42")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame(
                {
                    "a": [None, 1, 2, 2, 3, 4],
                    "b42": [0, 1, 2, 2, None, None],
                }
            ),
            check_row_order=False,
        )

        # test self-join
        assert_equal(
            tbl_left
            >> inner_join(
                tbl_left2 := tbl_left >> alias(),
                tbl_left.a == tbl_left2.a,
            ),
            df_left.join(df_left, on="a", coalesce=False, suffix="_df_left"),
        )

        assert_equal(
            tbl_right
            >> inner_join(tbl_right2 := tbl_right >> alias(), tbl_right.b == tbl_right2.b)
            >> inner_join(tbl_right3 := tbl_right >> alias(), tbl_right.b == tbl_right3.b),
            df_right.join(df_right, "b", suffix="_df_right", coalesce=False).join(
                df_right, "b", suffix="_df_right_1", coalesce=False
            ),
        )

    def test_filter(self, tbl1, tbl2):
        # Simple filter expressions
        assert_equal(tbl1 >> filter(tbl1.col1 == tbl1.col1), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == 3), df1.filter(pl.col("col1") == 3))

        # More complex expressions
        assert_equal(
            tbl1 >> filter(tbl1.col1 // 2 == 1),
            pl.DataFrame({"col1": [2, 3], "col2": ["b", "c"]}),
        )

        assert_equal(
            tbl1 >> filter(1 < tbl1.col1) >> filter(tbl1.col1 < 4),
            df1.filter((1 < pl.col("col1")) & (pl.col("col1") < 4)),
        )

    def test_arrange(self, tbl2, tbl4):
        tbl4.col1.nulls_first()

        assert_equal(
            tbl2 >> arrange(tbl2.col3) >> select(tbl2.col3),
            df2.select(pl.col("col3")).sort("col3", descending=False),
        )

        assert_equal(
            tbl2 >> arrange(-tbl2.col3) >> select(tbl2.col3),
            df2.select(pl.col("col3")).sort("col3", descending=True),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            df2.sort(["col1", "col2"], descending=[False, False]),
        )

        assert_equal(
            tbl4
            >> arrange(
                tbl4.col1.nulls_first(),
                tbl4.col2.nulls_first().descending(),
                tbl4.col5.nulls_first().descending(),
            ),
            df4.sort(
                ["col1", "col2", "col5"],
                descending=[False, True, True],
                nulls_last=False,
            ),
        )

        assert_equal(
            tbl4
            >> arrange(
                tbl4.col1.nulls_last(),
                tbl4.col2.descending().nulls_last(),
                tbl4.col5.descending().nulls_last(),
            ),
            df4.sort(
                ["col1", "col2", "col5"],
                descending=[False, True, True],
                nulls_last=True,
            ),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            tbl2 >> arrange(tbl2.col2) >> arrange(tbl2.col1),
        )

        assert_equal(tbl2 >> arrange(--tbl2.col3), tbl2 >> arrange(tbl2.col3))  # noqa: B002

    def test_summarize(self, tbl3):
        assert_equal(
            tbl3 >> summarize(mean=tbl3.col1.mean(), max=tbl3.col4.max()),
            pl.DataFrame({"mean": [1.0], "max": [11]}),
            check_row_order=False,
        )

        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarize(mean=tbl3.col4.mean()),
            pl.DataFrame({"col1": [0, 1, 2], "mean": [1.5, 5.5, 9.5]}),
            check_row_order=False,
        )

        assert_equal(
            tbl3 >> summarize(mean=tbl3.col4.mean()) >> mutate(mean_2x=C.mean * 2),
            pl.DataFrame({"mean": [5.5], "mean_2x": [11.0]}),
            check_row_order=False,
        )

    def test_group_by(self, tbl3):
        # Grouping doesn't change the result
        assert_equal(tbl3 >> group_by(tbl3.col1), tbl3)
        assert_equal(
            tbl3 >> summarize(mean4=tbl3.col4.mean()) >> group_by(C.mean4),
            tbl3 >> summarize(mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

        # Groupings can be added
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> group_by(tbl3.col2, add=True)
            >> summarize(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            tbl3 >> group_by(tbl3.col1, tbl3.col2) >> summarize(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

        # Ungroup doesn't change the result
        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarize(mean4=tbl3.col4.mean()) >> ungroup(),
            tbl3 >> group_by(tbl3.col1) >> summarize(mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

    def test_alias(self, tbl1, tbl2):
        x = tbl2 >> alias("x")
        assert x._ast.name == "x"

        # Check that applying alias doesn't change the output
        a = (
            tbl1
            >> mutate(xyz=(tbl1.col1 * tbl1.col1) // 2)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left", suffix="_right")
            >> mutate(col1=tbl1.col1 - C.xyz)
        )
        b = a >> alias("b")

        assert_equal(a, b)

        # self join
        assert_equal(
            tbl2 >> join(x, tbl2.col1 == x.col1, "left", suffix="_right"),
            df2.join(
                df2,
                how="left",
                left_on="col1",
                right_on="col1",
                coalesce=False,
            ),
        )

    def test_window_functions(self, tbl3, tbl4):
        # Everything else should stay the same
        assert_equal(
            tbl3 >> mutate(x=row_number(arrange=[-C.col4])) >> select(*tbl3),
            df3,
        )

        assert_equal(
            (tbl3 >> group_by(C.col2) >> mutate(x=row_number(arrange=[-C.col4])) >> select(C.x)),
            pl.DataFrame({"x": [6, 5, 6, 5, 4, 3, 4, 3, 2, 1, 2, 1]}),
        )

        # group_by and partition_by should lead to the same result
        assert_equal(
            (tbl3 >> group_by(C.col2) >> mutate(x=row_number(arrange=[-C.col4])) >> select(C.x)),
            (tbl3 >> mutate(x=row_number(arrange=[-C.col4], partition_by=[C.col2])) >> select(C.x)),
        )

        assert_equal(
            tbl3 >> mutate(x=tbl3.col1.shift(1, arrange=tbl3.col4)) >> inner_join(tbl4, on="col1"),
            df3.sort(pl.col("col4"))
            .with_columns(x=pl.col("col1").shift(1))
            .join(df4, on="col1", suffix="_df4", coalesce=False),
        )

    def test_slice_head(self, tbl3):
        @pdt.verb
        def slice_head_custom(table: pdt.Table, n: int, *, offset: int = 0):
            t = (
                table
                >> mutate(_n=row_number(arrange=[]))
                >> alias()
                >> filter((offset < C._n) & (C._n <= (n + offset)))
            )
            return t >> select(*[C[col.name] for col in table if col.name != "_n"])

        assert_equal(
            tbl3 >> slice_head(6),
            tbl3 >> slice_head_custom(6),
        )

        assert_equal(
            tbl3 >> slice_head(3, offset=2),
            tbl3 >> slice_head_custom(3, offset=2),
        )

        assert_equal(
            tbl3 >> slice_head(1, offset=100),
            tbl3 >> slice_head_custom(1, offset=100),
        )

    def test_case_expression(self, tbl3):
        assert_equal(
            (
                tbl3
                >> mutate(
                    col1=when(C.col1 == 0).then(1).when(C.col1 == 1).then(2).when(C.col1 == 2).then(3).otherwise(-1)
                )
                >> select(C.col1)
            ),
            (df3.select("col1") + 1),
        )

        assert_equal(
            (
                tbl3
                >> mutate(x=when(C.col1 == C.col2).then(1).when(C.col1 == C.col3).then(2).otherwise(C.col4))
                >> select(C.x)
            ),
            pl.DataFrame({"x": [1, 1, 2, 3, 4, 2, 1, 1, 8, 9, 2, 11]}),
        )

    def test_lambda_column(self, tbl1, tbl2):
        # Select
        assert_equal(tbl1 >> select(C.col1), tbl1 >> select(tbl1.col1))

        # Mutate
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> mutate(b=C.a * 2) >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> mutate(b=C.a * 2, a=tbl1.col1) >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        # Join
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1) >> join(tbl2, C.a == tbl2.col1, "left") >> select(C.a, *tbl2),
            tbl1 >> select() >> mutate(a=tbl1.col1) >> join(tbl2, tbl1.col1 == tbl2.col1, "left", suffix="_df2"),
        )

        # Filter
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter(C.a % 2 == 0),
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter((tbl1.col1 * 2) % 2 == 0),
        )

        # Arrange
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> arrange(C.a),
            tbl1 >> arrange(tbl1.col1) >> mutate(a=tbl1.col1 * 2),
        )

    def test_custom_verb(self, tbl1):
        @pdt.verb
        def double_col1(table):
            table >>= mutate(col1=C.col1 * 2)
            return table

        assert_equal(tbl1 >> double_col1(), tbl1 >> mutate(col1=C.col1 * 2))

    def test_null(self, tbl4):
        # see if we can compare to null
        assert_equal(
            tbl4 >> mutate(u=tbl4.col1 == tbl4.col3),
            df4.with_columns((pl.col("col1") == pl.col("col3")).alias("u")),
        )

        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.is_null()),
            df4.with_columns(pl.col("col3").is_null().alias("u")),
        )

        # test fill_null
        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.fill_null(tbl4.col2)),
            df4.with_columns(pl.col("col3").fill_null(pl.col("col2")).alias("u")),
        )
        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.fill_null(tbl4.col2)),
            tbl4 >> mutate(u=pdt.when(tbl4.col3.is_null()).then(tbl4.col2).otherwise(tbl4.col3)),
        )

    def test_datetime(self, tbl_dt):
        assert_equal(
            tbl_dt
            >> mutate(
                u=(tbl_dt.dt1 - tbl_dt.dt2),
                v=tbl_dt.d1 - tbl_dt.d1,
                w=(tbl_dt.d1.cast(pdt.Datetime) - tbl_dt.dt2) + tbl_dt.dur1 + dt.timedelta(days=1),
            ),
            df_dt.with_columns(
                (pl.col("dt1") - pl.col("dt2")).alias("u"),
                pl.duration().alias("v"),
                ((pl.col("d1") - pl.col("dt2")) + pl.col("dur1") + pl.lit(dt.timedelta(days=1))).alias("w"),
            ),
        )

    def test_duckdb_execution(self, tbl2, tbl3):
        assert_equal(
            tbl3 >> mutate(u=tbl3.col1 * 2) >> collect(DuckDb()) >> mutate(v=tbl3.col3 + C.u),
            tbl3 >> mutate(u=tbl3.col1 * 2) >> mutate(v=C.col3 + C.u),
        )

        assert_equal(
            tbl3
            >> collect(DuckDb())
            >> left_join(tbl2 >> collect(DuckDb()), tbl3.col1 == tbl2.col1, suffix="_right")
            >> mutate(v=tbl3.col3 + tbl2.col2)
            >> group_by(C.col2)
            >> summarize(y=C.col3_right.sum()),
            tbl3
            >> left_join(tbl2, tbl3.col1 == tbl2.col1, suffix="_right")
            >> mutate(v=C.col3 + C.col2_right)
            >> group_by(C.col2)
            >> summarize(y=C.col3_right.sum()),
            check_row_order=False,
        )

    def test_col_export(self, tbl1: pdt.Table, tbl2: pdt.Table):
        assert_equal(df1.get_column("col1"), tbl1.col1.export(Polars()))
        t = tbl2 >> mutate(u=tbl2.col1 * tbl2.col2, v=-tbl2.col3)
        t_ex: pl.DataFrame = t >> export(Polars(lazy=False))
        assert_equal(
            (t_ex["u"] + t_ex["col2"]).exp() - t_ex["v"],
            (t >> mutate(h=(t.u + C.col2).exp() - t.v)).h.export(Polars()),
        )
        assert_equal(
            (t_ex["u"] + t_ex["col2"]).exp() - t_ex["v"],
            ((t.u + C.col2).exp() - t.v).export(Polars()),
        )

        e = t >> inner_join(tbl1, tbl1.col1.cast(pdt.Float64()) <= tbl2.col1 + tbl2.col3)
        e_ex = e >> export(Polars(lazy=False))

        assert_equal(
            e_ex["col2"] + e_ex["col1_df1"],
            (e >> mutate(j=e.col2 + tbl1.col1)).j.export(Polars()),
        )
        assert_equal(
            e_ex["col2"] + e_ex["col1_df1"],
            (e.col2 + tbl1.col1).export(Polars),
        )

    def test_list(self, tbl1, tbl3):
        df = tbl1 >> export(Polars())
        df = df.with_columns(l=[[1, 2], [], [4, 5, 5], [2, 3]])
        assert_equal(pdt.Table(df), df)
        d = {"a": [["b", "aa"], ["a"], []]}
        t = pdt.Table(d)
        s = pl.DataFrame(d)
        assert_equal(t, s)
        assert_equal(
            t >> mutate(b=[[5], [0.3, 3.66], []]),
            s.with_columns(b=[[5], [0.3, 3.66], []]),
        )

        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarize(x=tbl3.col3.list.agg(arrange="col4")) >> arrange(tbl3.col1),
            df3.group_by(pl.col("col1")).agg(x=pl.col("col3").sort_by("col4")).sort("col1"),
        )

    def test_cum_sum(self, tbl1):
        assert_equal(
            tbl1 >> mutate(p=tbl1.col1.cum_sum(arrange=tbl1.col1)),
            df1.with_columns(p=pl.col("col1").cum_sum()),
        )

    def test_collect(self, tbl1):
        assert_equal(
            tbl1 >> collect() >> mutate(x=tbl1.col1),
            tbl1 >> mutate(x=tbl1.col1),
        )

        with pytest.raises(ColumnNotFoundError):
            tbl1 >> select() >> collect() >> mutate(x=tbl1.col1)

    def test_scalar_export(self):
        assert (pdt.Table({"a": 1}) >> export(pdt.Scalar)) == 1

    def test_dict_export(self):
        assert (pdt.Table({"a": 1}) >> export(pdt.Dict)) == {"a": 1}
        assert (pdt.Table({"a": 1, "b": True}) >> export(pdt.Dict)) == {
            "a": 1,
            "b": True,
        }

        with pytest.raises(TypeError):
            pdt.Table({"a": [1, 2]}) >> export(pdt.Dict)

    def test_dict_of_lists_export(self):
        assert (pdt.Table({"a": 1}) >> export(pdt.DictOfLists)) == {"a": [1]}
        assert (pdt.Table({"a": [1, 2], "b": [True, False]}) >> export(pdt.DictOfLists)) == {
            "a": [1, 2],
            "b": [True, False],
        }

    def test_list_of_dicts_export(self):
        assert (pdt.Table({"a": 1}) >> export(pdt.ListOfDicts)) == [{"a": 1}]
        assert (pdt.Table({"a": [1, 2], "b": [True, False]}) >> export(pdt.ListOfDicts)) == [
            {"a": 1, "b": True},
            {"a": 2, "b": False},
        ]

    def test_uses_table(self, tbl2, tbl3):
        assert tbl2.col1.uses_table(tbl2)
        assert not tbl2.col1.uses_table(tbl3)
        assert (tbl2.col1 == tbl3.col1).uses_table(tbl3)
        assert not tbl2.col1.uses_table(tbl2 >> mutate(x=0))

    def test_name(self, tbl3):
        assert tbl3 >> name() == "df3"

    def test_columns(self, tbl3):
        assert tbl3 >> columns() == [col.name for col in tbl3]
        tbl3_different_col_order = tbl3 >> mutate(col3=tbl3.col3)
        assert tbl3_different_col_order >> columns() == [col.name for col in tbl3_different_col_order]

    def test_name_alias(self, tbl2):
        assert tbl2 >> alias("tbl") >> name() == "tbl"

    def test_enum(self, tbl1):
        tbl1_enum = tbl1 >> mutate(p=tbl1.col2.cast(pdt.Enum("a", "b", "c", "d")))
        df1_enum = df1.with_columns(p=pl.col("col2").cast(pl.Enum(["a", "b", "c", "d"])))
        assert_equal(tbl1_enum, df1_enum)

        with pytest.raises(pl.exceptions.InvalidOperationError):
            (tbl1 >> mutate(p=tbl1.col2.cast(pdt.Enum("a", "b", "d"))) >> export(Polars))

        assert_equal(
            tbl1_enum >> mutate(q=C.p + "l"),
            df1_enum.with_columns(q=pl.col("p") + "l"),
        )

    def test_enum_isin(self, tbl1):
        tbl1 >> mutate(p=tbl1.col2.cast(pdt.Enum("a", "b", "c", "d")).is_in("a", "b", "c", "d"))
        tbl1 >> mutate(q="a") >> mutate(p=C.q.cast(pdt.Enum("a")).is_in("a"))

    def test_col_rename(self, tbl2, tbl4):
        assert_equal(tbl4 >> rename({tbl4.col1: "s"}), df4.rename({"col1": "s"}))
        assert_equal(
            tbl2
            >> left_join(tbl4, on=tbl2.col1 == tbl4.col1, suffix="_right")
            >> rename({tbl4.col5: "col5", tbl4.col4: "col4"})
            >> drop(tbl4.col1),
            df2.join(df4, how="left", on="col1"),
        )

    def test_join_renaming(self, tbl2, tbl3):
        assert_equal(
            tbl2 >> rename({"col3": "c"}) >> left_join(tbl3, on=["col1", "col2"]),
            df2.rename({"col3": "c"})
            .join(df3, how="left", on=["col1", "col2"])
            .with_columns(
                col1_df3=pl.when(pl.col("col4").is_null()).then(None).otherwise(pl.col("col1")),
                col2_df3=pl.when(pl.col("col4").is_null()).then(None).otherwise(pl.col("col2")),
            ),
        )

    def test_collision_renaming(self, tbl2, tbl3):
        with pytest.raises(ValueError, match="rename would cause duplicate column name `col3`"):
            _ = (
                tbl2
                >> mutate(c=tbl2.col3)
                >> inner_join(
                    tbl3 >> select(),
                    (tbl2.col1 == tbl3.col1) & (tbl2.col2 == tbl3.col2),
                )
                >> rename({"c": "col3"}),
                df2.join(df3, how="inner", on=["col1", "col2"]).select("col1", "col2", "col3"),
            )

    def test_hidden_collision_renaming0(self, tbl2, tbl3):
        assert_equal(
            tbl2
            >> rename({"col3": "c"})
            >> inner_join(tbl3 >> select(), (tbl2.col1 == tbl3.col1) & (tbl2.col2 == tbl3.col2))
            >> rename({"c": "col3"}),
            df2.join(df3, how="inner", on=["col1", "col2"]).select("col1", "col2", "col3"),
        )

    def test_hidden_collision_renaming1(self, tbl2, tbl3):
        assert_equal(
            tbl2
            >> rename({"col3": "c"})
            >> collect()
            >> inner_join(tbl3 >> select(), (tbl2.col1 == tbl3.col1) & (tbl2.col2 == tbl3.col2))
            >> rename({C.c: "col3"}),
            df2.join(df3, how="inner", on=["col1", "col2"]).select("col1", "col2", "col3"),
        )

    def test_hidden_collision_renaming2(self, tbl2, tbl3):
        assert_equal(
            tbl2
            >> rename({"col3": "c"})
            >> inner_join(tbl3 >> select(), (tbl2.col1 == tbl3.col1) & (tbl2.col2 == tbl3.col2))
            >> mutate(col3=C.c),
            df2.join(df3, how="inner", on=["col1", "col2"])
            .select("col1", "col2", "col3")
            .with_columns(c=pl.col("col3")),
        )

    def test_hidden_collision_renaming3(self, tbl2, tbl3):
        assert_equal(
            tbl2
            >> rename({"col3": "c"})
            >> collect()
            >> inner_join(tbl3 >> select(), (tbl2.col1 == tbl3.col1) & (tbl2.col2 == tbl3.col2))
            >> mutate(col3=C.c),
            df2.join(df3, how="inner", on=["col1", "col2"])
            .select("col1", "col2", "col3")
            .with_columns(c=pl.col("col3")),
        )

    def test_eval_aligned(self, tbl4):
        assert_equal(
            tbl4,
            tbl4 >> drop(tbl4.col1) >> mutate(col1=eval_aligned(df4.get_column("col1"))),
        )

        s = pl.Series([1.2**i for i in range(df4.height)])
        assert_equal(
            tbl4 >> filter(tbl4.col3 <= eval_aligned(s, with_=tbl4)),
            df4.filter(pl.col("col3") <= s),
        )

        with pytest.raises(TypeError):
            tbl4 >> arrange(s)

        # pandas series
        s_pd = s.to_pandas()
        assert_equal(
            tbl4 >> mutate(z=eval_aligned(tbl4.col1 + s_pd)) >> group_by(tbl4.col3) >> summarize(u=C.z.sum()),
            df4.with_columns(z=pl.col("col1") + s).group_by("col3").agg(u=pl.col("z").sum()),
            check_row_order=False,
        )

    def test_aligned_decorator(self, tbl3):
        @aligned(with_="tbl")
        def reverse_col(col: pdt.ColExpr, tbl: pdt.Table) -> pdt.ColExpr:
            pl_col: pl.Series = col.export(Polars())
            return pl_col.reverse()

        assert_equal(
            tbl3 >> mutate(r=reverse_col(tbl3.col1 + tbl3.col2, tbl3)),
            df3.with_columns(r=(pl.col("col1") + pl.col("col2")).reverse()),
        )

    def test_aligned_col_export(self, tbl3):
        s = pl.Series(list(2**i for i in range(df3.height)))
        assert_equal(
            eval_aligned(tbl3.col1 + s, with_=tbl3).export(Polars),
            df3.get_column("col1") + s,
        )
        assert_equal(
            eval_aligned(pdt.lit(3) * s, with_=tbl3).export(Polars),
            s * 3,
        )

    def test_aligned_multiple_tables(self, tbl1, tbl_right):
        assert_equal(
            tbl1 >> mutate(z=eval_aligned(tbl1.col1 + tbl_right.b)),
            df1.with_columns(z=pl.col("col1") + df_right.get_column("b")),
        )

    def test_rand(self, tbl4):
        tbl4 >> mutate(r=pdt.rand()) >> export(Polars)


class TestPrintAndRepr:
    def test_table_str(self, tbl1):
        tbl_str = str(tbl1)

        assert "df1" in tbl_str
        assert "polars" in tbl_str
        assert str(df1) in tbl_str
        tbl1 >> show()

    def test_table_repr_html(self, tbl1):
        # jupyter html
        assert "failed" not in tbl1._repr_html_()
        assert tbl1._repr_html_() == df1._repr_html_()

    def test_col_str(self, tbl1):
        col1_str = str(tbl1.col1)
        col_tbl = tbl1._ast.df.select("col1").collect()

        assert str(col_tbl).split("\n", 1)[1] in col1_str
        assert "failed" not in col1_str

    def test_col_html_repr(self, tbl1):
        assert "failed" not in tbl1.col1._repr_html_()

    def test_expr_str(self, tbl1):
        expr_str = str(tbl1.col1 * 2)
        logger.info(expr_str)
        assert "failed" not in expr_str

    def test_expr_html_repr(self, tbl1):
        assert "failed" not in (tbl1.col1 * 2)._repr_html_()

    def test_preview_print(self, tbl3):
        tbl3_str = str(tbl3)
        assert "failed" not in tbl3_str
        assert "shape: (12, 5)" in tbl3_str

    def test_ast_repr(self, tbl4):
        tbl4.col1.ast_repr()
        (tbl4.col1 + tbl4.col2).ast_repr()
        (tbl4.col1 + tbl4.col2 + tbl4.col3).ast_repr()
        (pdt.when(tbl4.col1 > 1).then(tbl4.col2).when(tbl4.col1 < -1).then(tbl4.col3).otherwise(7)).ast_repr()

        (tbl4.col1.cast(pdt.Float64) + tbl4.col2 / 2).ast_repr()

        # TODO: This is currently how `filter` is translated to the AST. We could also
        # only do this transformation during backend translation, so that there is a
        # real `filter` context kwarg visible in the AST.

        tbl4.col1.max(
            partition_by=[tbl4.col2, tbl4.col3],
            filter=pdt.when(tbl4.col1 > 0).then(tbl4.col2.is_not_null()).otherwise((tbl4.col3 % 2) == 0),
        ).ast_repr()

    def test_error_source_ptr(self, tbl1, tbl2):
        with pytest.raises(DataTypeError) as r:
            tbl1 >> mutate(z=(2 * (tbl1.col1 + C.col2) - tbl1.col1.exp()) * tbl1.col1.acos())
        assert "AST path" in r.value.args[0]

        with pytest.raises(DataTypeError) as r:
            tbl1 >> inner_join(
                tbl2 >> rename({c: c.name + "42" for c in tbl2}),
                on=C.col2 == tbl2.col1,
            )
        assert "AST path" in r.value.args[0]

    def test_verb_ast_repr(self, tbl3, tbl4):
        intermed = (
            tbl3
            >> mutate(
                u=C.col1 + 5,
                v=(tbl3.col2.exp() + tbl3.col4) * tbl3.col1,
                w=pdt.max(tbl3.col2, tbl3.col1, tbl3.col5.str.len()),
                x=pdt.count(),
            )
            >> select(tbl3.col1, tbl3.col4)
            >> alias("tbl 42")
        )

        ast_str = (
            intermed
            >> left_join(
                tbl4
                >> arrange(tbl4.col4.descending())
                >> filter(tbl4.col5.str.len() <= 10)
                >> full_join(s := tbl4 >> alias("s"), on=s.col1 == tbl4.col1),
                on="col1",
            )
            >> slice_head(32)
            >> group_by()
            >> summarize(u=pdt.when(intermed.col1.max() >= 2).then(None).otherwise(4))
            >> ast_repr(verb_depth=-1, expr_depth=-1)
        )

        series = pl.Series([2**i for i in range(12)])

        ast_str = (
            tbl3
            >> mutate(
                u=C.col1 + 5,
                v=(tbl3.col2.exp() + tbl3.col4) * tbl3.col1,
                w=pdt.max(tbl3.col2, tbl3.col1, tbl3.col5.str.len()),
                x=pdt.count(),
                y=eval_aligned((tbl3 >> mutate(j=42) >> alias("tbl. 1729") >> filter(C.col2 > 0)).j + tbl3.col1),
                z=eval_aligned(series),
            )
            >> select(tbl3.col1, tbl3.col4)
            >> left_join(tbl4, on="col1")
        )._ast.ast_repr()

        logger.info(f"AST:\n{ast_str}")

        assert "tbl__1729.j" in ast_str



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from ._internal.pipe.c import C
from ._internal.pipe.verbs import alias, build_query, collect, export, show, show_query

__all__ = [
    "C",
    "alias",
    "build_query",
    "collect",
    "export",
    "show",
    "show_query",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import functools
import itertools
import operator
import uuid
from typing import Any, Literal, overload

import polars as pl

from pydiverse.common import Bool, Int64
from pydiverse.transform._internal import errors
from pydiverse.transform._internal.backend.table_impl import (
    TableImpl,
    split_join_cond,
)
from pydiverse.transform._internal.backend.targets import (
    Dict,
    DictOfLists,
    ListOfDicts,
    Polars,
    Scalar,
    Target,
)
from pydiverse.transform._internal.errors import (
    ColumnNotFoundError,
    DataTypeError,
    ErrorWithSource,
    FunctionTypeError,
)
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.pipe.pipeable import (
    Pipeable,
    check_subquery,
    modify_ast,
    verb,
)
from pydiverse.transform._internal.pipe.table import Table
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.col_expr import (
    Col,
    ColExpr,
    ColFn,
    ColName,
    EvalAligned,
    LiteralCol,
    Order,
    Series,
    get_ast_path_str,
    wrap_literals,
)
from pydiverse.transform._internal.tree.verbs import (
    Alias,
    Arrange,
    Filter,
    GroupBy,
    Join,
    Mutate,
    Rename,
    Select,
    SliceHead,
    Summarize,
    Ungroup,
    Union,
)

__all__ = [
    "alias",
    "collect",
    "build_query",
    "show_query",
    "select",
    "drop",
    "rename",
    "mutate",
    "join",
    "left_join",
    "inner_join",
    "full_join",
    "union",
    "filter",
    "arrange",
    "group_by",
    "ungroup",
    "summarize",
    "slice_head",
    "export",
]


@overload
def alias(new_name: str | None = None, *, keep_col_refs: bool = False) -> Pipeable: ...


@verb
@modify_ast
def alias(table: Table, new_name: str | None = None, *, keep_col_refs: bool = False) -> Pipeable:
    """
    Changes the name of the current table and allows subqueries in SQL.

    :param new_name:
        The new name assigned to the table. If this is ``None``, the table
        retains its previous name.

    :param keep_col_refs:
        Whether column references are reset. If ``keep_col_refs=False``, the resulting
        table is not considered to be derived from the input table, i.e. one cannot use
        columns from the input table in subsequent operations on the result table.
        However, the reset of column references is necessary before performing a
        self-join.

    Examples
    --------

    A self join without applying ``alias`` before raises an exception:

    >>> t = pdt.Table({"a": [4, 2, 1, 4], "b": ["l", "g", "uu", "--   r"]})
    >>> t >> join(t, t.a == t.a, how="inner", suffix="_right")
    ValueError: table `<pydiverse.transform._internal.backend.polars.PolarsImpl object
    at 0x13f86d510>` occurs twice in the table tree.
    hint: To join two tables derived from a common table, apply `>> alias()` to one of
    them before the join.

    By applying ``alias`` to the right table and storing the result in a new variable,
    the self join succeeds.

    >>> (
    ...     t
    ...     >> join(s := t >> alias(), t.a == s.a, how="inner", suffix="_right")
    ...     >> show()
    ... )
    shape: (6, 4)
    ┌─────┬────────┬─────────┬─────────┐
    │ a   ┆ b      ┆ a_right ┆ b_right │
    │ --- ┆ ---    ┆ ---     ┆ ---     │
    │ i64 ┆ str    ┆ i64     ┆ str     │
    ╞═════╪════════╪═════════╪═════════╡
    │ 4   ┆ l      ┆ 4       ┆ l       │
    │ 4   ┆ --   r ┆ 4       ┆ l       │
    │ 2   ┆ g      ┆ 2       ┆ g       │
    │ 1   ┆ uu     ┆ 1       ┆ uu      │
    │ 4   ┆ l      ┆ 4       ┆ --   r  │
    │ 4   ┆ --   r ┆ 4       ┆ --   r  │
    └─────┴────────┴─────────┴─────────┘
    """

    if new_name is None:
        new_name = table._ast.name

    new = copy.copy(table)
    new._ast = Alias(
        new._ast,
        uuid_map={uid: uuid.uuid1() for uid in table._cache.cols.keys()} if not keep_col_refs else None,
    )
    new._ast.name = new_name

    return new


@overload
def collect(target: Target | None = None, *, keep_col_refs: bool = True) -> Pipeable: ...


@verb
def collect(
    table: Table,
    target: Target | None = None,
    *,
    keep_col_refs: bool = True,
) -> Pipeable:
    """
    Execute all accumulated operations and write the result to a new Table.

    This verb is only for polars-backed tables. All operations lazily stored in the
    table are executed and a new table containing the result is returned. The returned
    table always stored the data in a polars LazyFrame. One can choose whether the
    following operations on the table are executed via polars or DuckDB on the
    LazyFrame (see also :doc:`/examples/duckdb_polars_parquet`).

    :param target:
        The execution engine to be used from here on. Can be either ``Polars`` or
        ``DuckDb``.

    Examples
    --------
    Here, ``collect`` does not change anything in the result, but the ``mutate`` is
    executed on the DataFrame when ``collect`` is called, whereas the ``arrange`` is
    only executed when ``export`` is called. Without ``collect``, the ``mutate`` would
    only have been executed with the ``export``, too.

    >>> t = pdt.Table({"a": [4, 2, 1, 4], "b": ["l", "g", "uu", "--   r"]})
    >>> (
    ...     t
    ...     >> mutate(z=t.a + t.b.str.len())
    ...     >> collect()
    ...     >> arrange(C.z, t.a)
    ...     >> show()
    ... )
    shape: (4, 3)
    ┌─────┬────────┬─────┐
    │ a   ┆ b      ┆ z   │
    │ --- ┆ ---    ┆ --- │
    │ i64 ┆ str    ┆ i64 │
    ╞═════╪════════╪═════╡
    │ 1   ┆ uu     ┆ 3   │
    │ 2   ┆ g      ┆ 3   │
    │ 4   ┆ l      ┆ 5   │
    │ 4   ┆ --   r ┆ 10  │
    └─────┴────────┴─────┘
    """
    errors.check_arg_type(Target | type(Target) | type(None), "collect", "target", target)

    if target is None:
        target = Polars()

    if not isinstance(target, Target):
        assert issubclass(target, Target)
        target = target()

    df = table >> export(Polars(lazy=False))

    if not keep_col_refs:
        return Table(df)

    # TODO: keep_hidden_cols option

    assert len(table) == len(table._cache.name_to_uuid)

    new = Table(
        TableImpl.from_resource(
            df,
            target,
            name=table._ast.name,
            # preserve UUIDs -> this keeps column references across collect()
            uuids={name: uid for name, uid in table._cache.name_to_uuid.items()},
        )
    )
    new._cache.derived_from = table._cache.derived_from | {new._ast}
    new._cache.partition_by = [preprocess_arg(col, new) for col in table._cache.partition_by]

    return new


@overload
def export(target: Target, *, schema_overrides: dict[str, Any] | None = None) -> Pipeable: ...


@verb
def export(
    table: Table,
    target: Target | type[Target],
    *,
    schema_overrides: dict[str, Any] | None = None,
) -> Pipeable:
    """Convert a pydiverse.transform Table to a data frame.

    :param target:
        Can currently be either a ``Polars`` or ``Pandas`` object. For polars, one can
        specify whether a DataFrame or LazyFrame is returned via the ``lazy`` keyword
        parameter.
        If ``lazy=True``, no actual computations are performed, they just get stored in
        the LazyFrame.

    :param schema_overrides:
        A dictionary of column names to backend-specific data types. This controls which
        data types are used when writing to the backend. Because the data types are not
        constrained by pydiverse.transform's type system, this may sometimes be
        preferred over a cast.

    :return:
        A polars or pandas DataFrame / LazyFrame.

    Examples
    --------

    >>> t1 = pdt.Table(
    ...     {
    ...         "a": [3, 1, 4, 1, 5, 9],
    ...         "b": [2.465, 0.22, -4.477, 10.8, -81.2, 0.0],
    ...         "c": ["a,", "bcbc", "pydiverse", "transform", "'  '", "-22"],
    ...     }
    ... )
    >>> t1 >> export(Pandas())
       a      b          c
    0  3  2.465         a,
    1  1   0.22       bcbc
    2  4 -4.477  pydiverse
    3  1   10.8  transform
    4  5  -81.2       '  '
    5  9    0.0        -22
    >>> t1 >> show()
    shape: (6, 3)
    ┌─────┬────────┬───────────┐
    │ a   ┆ b      ┆ c         │
    │ --- ┆ ---    ┆ ---       │
    │ i64 ┆ f64    ┆ str       │
    ╞═════╪════════╪═══════════╡
    │ 3   ┆ 2.465  ┆ a,        │
    │ 1   ┆ 0.22   ┆ bcbc      │
    │ 4   ┆ -4.477 ┆ pydiverse │
    │ 1   ┆ 10.8   ┆ transform │
    │ 5   ┆ -81.2  ┆ '  '      │
    │ 9   ┆ 0.0    ┆ -22       │
    └─────┴────────┴───────────┘
    """

    errors.check_arg_type(Target | type(Target), "export", "target", target)

    if not isinstance(target, Target):
        assert issubclass(target, Target)
        target = target()

    if isinstance(target, Scalar):
        if len(table) != 1:
            raise TypeError(
                f"to export a table to a scalar, it must have exactly one column, but found {len(table)} columns"
            )
        df: pl.DataFrame = table >> export(Polars())
        if df.height != 1:
            raise TypeError(f"to export a table to a scalar, it must have exactly one row, but found {df.height} rows")
        return df.item()

    elif isinstance(target, Dict):
        df: pl.DataFrame = table >> export(Polars())
        if df.height != 1:
            raise TypeError(f"cannot export a table with more than one row to `Dict`, found {df.height} rows")
        return df.to_dicts()[0]

    elif isinstance(target, DictOfLists):
        return (table >> export(Polars())).to_dict(as_series=False)

    elif isinstance(target, ListOfDicts):
        return (table >> export(Polars())).to_dicts()

    # TODO: allow stuff like pdt.Int(): pl.UInt32() in schema_overrides and resolve that
    # to columns
    SourceBackend: type[TableImpl] = table._cache.backend
    if schema_overrides is None:
        schema_overrides = dict()
    return SourceBackend.export(
        table._ast.clone(),
        target,
        schema_overrides={table[col_name]._uuid: dtype for col_name, dtype in schema_overrides.items()},
    )


@overload
def build_query() -> Pipeable: ...


@verb
def build_query(table: Table) -> str | None:
    """
    Compiles the operations accumulated on the current table to a SQL query.

    :returns:
        The SQL query as a string.
    """

    return table._cache.backend.build_query(table._ast.clone())


@overload
def show_query(pipe: bool = False) -> Pipeable | None: ...


@verb
def show_query(table: Table, pipe: bool = False) -> Pipeable | None:
    """
    Prints the compiled SQL query to stdout.

    :param pipe:
        If set to `True`, the table is returned, else `None` is returned.

    Note
    ----
    During interactive development in the python shell, you usually want to keep
    `pipe = False`, else the table is printed, too. The main use for `pipe = True` is
    when you don't want to break a long sequence of verbs in a file.
    """

    if query := table >> build_query():
        print(query)
    else:
        print(f"No query to show for table {table._ast.short_name()}. (backend: {table._cache.backend.backend_name})")

    return table if pipe else None


@overload
def select(*cols: Col | ColName | str) -> Pipeable: ...


@verb
@modify_ast
def select(table: Table, *cols: Col | ColName | str) -> Pipeable:
    """
    Selects a subset of columns.

    :param cols:
        The columns to be included in the resulting table.

    Examples
    --------
    >>> t = pdt.Table({"a": [3, 2, 6, 4], "b": ["lll", "g", "u0", "__**_"]})
    >>> t >> select(t.a) >> show()
    shape: (4, 1)
    ┌─────┐
    │ a   │
    │ --- │
    │ i64 │
    ╞═════╡
    │ 3   │
    │ 2   │
    │ 6   │
    │ 4   │
    └─────┘
    """

    errors.check_vararg_type(Col | ColName | str, "select", *cols)

    for col in cols:
        if isinstance(col, ColName | str) and col not in table:
            raise ColumnNotFoundError(f"column `{col.ast_repr()}` does not exist in table `{table._ast.short_name()}`")
        elif col not in table and col._uuid in table._cache.cols:
            raise ColumnNotFoundError(
                f"cannot select hidden column `{col.ast_repr()}` again\n"
                "hint: A column becomes hidden if you deselected it before or "
                "overwrite it in `mutate` or `summarize`."
            )

    cols = [ColName(col) if isinstance(col, str) else col for col in cols]

    new = copy.copy(table)
    new._ast = Select(table._ast, [preprocess_arg(col, table) for col in cols])

    return new


@overload
def drop(*cols: Col | ColName | str) -> Pipeable: ...


@verb
def drop(table: Table, *cols: Col | ColName | str) -> Pipeable:
    """
    Removes a subset of the columns.

    :param cols:
        The columns to be removed.

    Examples
    --------
    >>> t = pdt.Table({"a": [3, 2, 6, 4], "b": ["lll", "g", "u0", "__**_"]})
    >>> t >> drop(t.a) >> show()
    shape: (4, 1)
    ┌───────┐
    │ b     │
    │ ---   │
    │ str   │
    ╞═══════╡
    │ lll   │
    │ g     │
    │ u0    │
    │ __**_ │
    └───────┘
    """
    errors.check_vararg_type(Col | ColName | str, "drop", *cols)
    cols = [ColName(col) if isinstance(col, str) else col for col in cols]

    dropped_uuids = {preprocess_arg(col, table)._uuid for col in cols}
    return table >> select(
        *(name for name, uid in table._cache.name_to_uuid.items() if uid not in dropped_uuids),
    )


@overload
def rename(name_map: dict[str, str]) -> Pipeable: ...


@verb
@modify_ast
def rename(table: Table, name_map: dict[str | Col | ColName, str]) -> Pipeable:
    """
    Renames columns.

    :param name_map:
        A dictionary assigning some columns new names.

    Examples
    --------
    Renaming one column:

    >>> t = pdt.Table({"a": [3, 2, 6, 4], "b": ["lll", "g", "u0", "__**_"]})
    >>> t >> rename({"a": "h"}) >> show()
    shape: (4, 2)
    ┌─────┬───────┐
    │ h   ┆ b     │
    │ --- ┆ ---   │
    │ i64 ┆ str   │
    ╞═════╪═══════╡
    │ 3   ┆ lll   │
    │ 2   ┆ g     │
    │ 6   ┆ u0    │
    │ 4   ┆ __**_ │
    └─────┴───────┘

    Here is a more subtle example: As long as there are no two equal column names in the
    result table, one can give names to columns that already exist in the table. In the
    following example, the names of columns *a* and *b* are swapped.

    >>> s = t >> rename({"a": "b", "b": "a"}) >> show()
    Table <unnamed>, backend: PolarsImpl
    shape: (4, 2)
    ┌─────┬───────┐
    │ b   ┆ a     │
    │ --- ┆ ---   │
    │ i64 ┆ str   │
    ╞═════╪═══════╡
    │ 3   ┆ lll   │
    │ 2   ┆ g     │
    │ 6   ┆ u0    │
    │ 4   ┆ __**_ │
    └─────┴───────┘

    When using the column ``t.a`` in an expression derived from *s* now, it
    still refers to the same column, which now has the name *b*. The anonymous
    column ``C.a``, however, refers to the column with name *a* in the *current*
    table.

    >>> s >> mutate(u=t.a, v=C.a) >> show()
    shape: (4, 4)
    ┌─────┬───────┬─────┬───────┐
    │ b   ┆ a     ┆ u   ┆ v     │
    │ --- ┆ ---   ┆ --- ┆ ---   │
    │ i64 ┆ str   ┆ i64 ┆ str   │
    ╞═════╪═══════╪═════╪═══════╡
    │ 3   ┆ lll   ┆ 3   ┆ lll   │
    │ 2   ┆ g     ┆ 2   ┆ g     │
    │ 6   ┆ u0    ┆ 6   ┆ u0    │
    │ 4   ┆ __**_ ┆ 4   ┆ __**_ │
    └─────┴───────┴─────┴───────┘
    """
    errors.check_arg_type(dict, "rename", "name_map", name_map)

    for v in name_map.values():
        if not isinstance(v, str):
            raise TypeError(
                f"values in the `name_map` of `rename` must have type `str`, found `{v.__class__.__name__}` instead"
            )
    for k in name_map.keys():
        if not isinstance(k, str | Col | ColName):
            raise TypeError(
                "keys in the `name_map` of `rename` must have type `str | Col | "
                f"ColName`, found `{k.__class__.__name__}` instead"
            )

    name_map = {(preprocess_arg(k, table) if isinstance(k, ColName | Col) else k): v for k, v in name_map.items()}
    name_map = {(table._cache.uuid_to_name[k._uuid] if isinstance(k, Col) else k): v for k, v in name_map.items()}

    if d := set(name_map).difference(table._cache.name_to_uuid):
        raise ValueError(f"no column with name `{next(iter(d))}` in table `{table._ast.short_name()}`")

    if d := (set(table._cache.name_to_uuid).difference(name_map)) & set(name_map.values()):
        raise ValueError(f"rename would cause duplicate column name `{next(iter(d))}`")

    new = copy.copy(table)
    new._ast = Rename(table._ast, name_map)

    return new


@overload
def mutate(**kwargs: ColExpr) -> Pipeable: ...


@verb
@modify_ast
def mutate(table: Table, **kwargs: ColExpr) -> Pipeable:
    """
    Adds new columns to the table.

    :param kwargs:
        Each key is the name of a new column, and its value is the column
        expression defining the new column.

    Examples
    --------
    >>> t1 = pdt.Table(
    ...     dict(a=[3, 1, 4, 1, 5, 9], b=[2.465, 0.22, -4.477, 10.8, -81.2, 0.0])
    ... )
    >>> t1 >> mutate(u=t1.a * t1.b) >> show()
    shape: (6, 3)
    ┌─────┬────────┬─────────┐
    │ a   ┆ b      ┆ u       │
    │ --- ┆ ---    ┆ ---     │
    │ i64 ┆ f64    ┆ f64     │
    ╞═════╪════════╪═════════╡
    │ 3   ┆ 2.465  ┆ 7.395   │
    │ 1   ┆ 0.22   ┆ 0.22    │
    │ 4   ┆ -4.477 ┆ -17.908 │
    │ 1   ┆ 10.8   ┆ 10.8    │
    │ 5   ┆ -81.2  ┆ -406.0  │
    │ 9   ┆ 0.0    ┆ 0.0     │
    └─────┴────────┴─────────┘
    """

    names, values = list(kwargs.keys()), list(kwargs.values())
    uuids = [uuid.uuid1() for _ in names]
    new = copy.copy(table)
    preprocessed = []
    for name, val in zip(names, values, strict=True):
        try:
            preprocessed.append(preprocess_arg(val, table))
        except ErrorWithSource as e:
            raise type(e)(
                e.args[0] + f"\noccurred in `mutate` argument `{name}`\n"
                f"AST path to error source: {get_ast_path_str(val, e.source)}"
            ) from e
    new._ast = Mutate(table._ast, names, preprocessed, uuids)

    return new


@overload
def filter(*predicates: ColExpr[Bool]) -> Pipeable: ...


@verb
@modify_ast
def filter(table: Table, *predicates: ColExpr[Bool]) -> Pipeable:
    """
    Selects a subset of rows based on some condition.

    :param predicates:
        Column expressions of boolean type to filter by. Only rows where all expressions
        are true are included in the result.

    Examples
    --------
    >>> t = pdt.Table({"a": [3, 2, 6, 4], "b": ["lll", "g", "u0", "__**_"]})
    >>> t >> filter(t.a <= 4, ~t.b.str.contains("_")) >> show()
    shape: (2, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    │ --- ┆ --- │
    │ i64 ┆ str │
    ╞═════╪═════╡
    │ 3   ┆ lll │
    │ 2   ┆ g   │
    └─────┴─────┘
    """
    new = copy.copy(table)
    preprocessed = []
    for i, pred in enumerate(predicates, 1):
        try:
            preprocessed.append(preprocess_arg(pred, table))
        except ErrorWithSource as e:
            raise type(e)(
                e.args[0] + f"\noccurred in {i}-{ordinal_suffix(i)} `filter` argument\n"
                f"AST path to error source: {get_ast_path_str(pred, e.source)}"
            ) from e

    new._ast = Filter(table._ast, preprocessed)

    for cond in new._ast.predicates:
        if not types.without_const(cond.dtype()) == Bool():
            raise DataTypeError(
                "predicates given to `filter` must be of boolean type.\n"
                f"hint: {cond} is of type {cond.dtype()} instead."
            )

        for fn in cond.iter_subtree_postorder():
            if isinstance(fn, ColFn) and fn.op.ftype in (
                Ftype.WINDOW,
                Ftype.AGGREGATE,
            ):
                raise FunctionTypeError(
                    f"forbidden window function `{fn.op.name}` in `filter`\nhint: If "
                    "you want to filter by an expression containing a window / "
                    "aggregation function, first add the expression as a column via "
                    "`mutate`."
                )

    return new


@overload
def arrange(*order_by: ColExpr) -> Pipeable: ...


@verb
@modify_ast
def arrange(table: Table, by: ColExpr, *more_by: ColExpr) -> Pipeable:
    """
    Sorts the rows of the table.

    :param order_by:
        Column expressions to sort by. The order of the expressions determines
        the priority.

    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "r": [2, 7, 3, 2, 6, None, 4],
    ...         "s": ["l", "o", "a", "c", "s", "---", "3"],
    ...         "p": [0.655, -4.33, None, 143.6, 0.0, 1.0, 4.5],
    ...     }
    ... )
    >>> t >> arrange(t.r.nulls_first(), t.p) >> show()
    shape: (7, 3)
    ┌──────┬─────┬───────┐
    │ r    ┆ s   ┆ p     │
    │ ---  ┆ --- ┆ ---   │
    │ i64  ┆ str ┆ f64   │
    ╞══════╪═════╪═══════╡
    │ null ┆ --- ┆ 1.0   │
    │ 2    ┆ l   ┆ 0.655 │
    │ 2    ┆ c   ┆ 143.6 │
    │ 3    ┆ a   ┆ null  │
    │ 4    ┆ 3   ┆ 4.5   │
    │ 6    ┆ s   ┆ 0.0   │
    │ 7    ┆ o   ┆ -4.33 │
    └──────┴─────┴───────┘
    >>> t >> arrange(t.p.nulls_last().descending(), t.s) >> show()
    shape: (7, 3)
    ┌──────┬─────┬───────┐
    │ r    ┆ s   ┆ p     │
    │ ---  ┆ --- ┆ ---   │
    │ i64  ┆ str ┆ f64   │
    ╞══════╪═════╪═══════╡
    │ 2    ┆ c   ┆ 143.6 │
    │ 4    ┆ 3   ┆ 4.5   │
    │ null ┆ --- ┆ 1.0   │
    │ 2    ┆ l   ┆ 0.655 │
    │ 6    ┆ s   ┆ 0.0   │
    │ 7    ┆ o   ┆ -4.33 │
    │ 3    ┆ a   ┆ null  │
    └──────┴─────┴───────┘
    """

    order_by = [ColName(col) if isinstance(col, str) else col for col in (by, *more_by)]
    new = copy.copy(table)

    preprocessed = []
    for i, ord in enumerate(order_by, 1):
        try:
            preprocessed.append(preprocess_arg(Order.from_col_expr(ord), table))
        except ErrorWithSource as e:
            raise type(e)(
                e.args[0] + f"\noccurred in {i}-{ordinal_suffix(i)} `arrange` argument\n"
                f"AST path to error source: {get_ast_path_str(ord, e.source)}"
            ) from e

    new._ast = Arrange(table._ast, preprocessed)

    return new


@overload
def group_by(table: Table, *cols: Col | ColName | str, add=False) -> Pipeable: ...


@verb
@modify_ast
def group_by(table: Table, *cols: Col | ColName | str, add=False) -> Pipeable:
    """
    Add a grouping state to the table.

    :param cols:
        The columns to group by.

    :param add:
        If ``add=True``, the given columns are added to the set of columns the table is
        currently grouped by. If ``add=False``, the current grouping state is replaced
        by the given columns.

    This verb does not modify the table itself, but only adds a grouping in the
    background. The number of rows is only reduced when
    :doc:`summarize <pydiverse.transform.summarize>`
    is called. The
    :doc:`ungroup <pydiverse.transform.ungroup>`
    verb can be used to clear the grouping state.
    """
    errors.check_vararg_type(Col | ColName | str, "group_by", *cols)
    errors.check_arg_type(bool, "group_by", "add", add)
    cols = [ColName(col) if isinstance(col, str) else col for col in cols]

    for col in cols:
        if isinstance(col, Col) and col._uuid not in table._cache.uuid_to_name:
            raise ValueError(f"cannot group by non-selected column `{col.ast_repr()}`")

    new = copy.copy(table)
    new._ast = GroupBy(table._ast, [preprocess_arg(col, table) for col in cols], add)

    return new


@overload
def ungroup() -> Pipeable: ...


@verb
@modify_ast
def ungroup(table: Table) -> Pipeable:
    """
    Clear the grouping state of the table.

    Examples
    --------
    In the following example, ``group_by`` and ``ungroup`` are used to specify that each
    aggregation function between them uses the column ``t.c`` for grouping.

    >>> t = pdt.Table(
    ...     {
    ...         "a": [1.2, 5.077, -2.29, -0.0, 3.0, -7.7],
    ...         "b": ["a  ", "transform", "pipedag", "cdegh", "  -ade ", "  pq"],
    ...         "c": [True, False, None, None, True, True],
    ...         "d": [4, 4, 2, 0, 1, 0],
    ...     }
    ... )
    >>> (
    ...     t
    ...     >> group_by(t.c)
    ...     >> mutate(
    ...         u=t.b.str.len().max() + t.a.min(),
    ...         v=t.d.mean(filter=t.a >= 0),
    ...     )
    ...     >> ungroup()
    ...     >> show()
    ... )
    shape: (6, 6)
    ┌───────┬───────────┬───────┬─────┬────────┬─────┐
    │ a     ┆ b         ┆ c     ┆ d   ┆ u      ┆ v   │
    │ ---   ┆ ---       ┆ ---   ┆ --- ┆ ---    ┆ --- │
    │ f64   ┆ str       ┆ bool  ┆ i64 ┆ f64    ┆ f64 │
    ╞═══════╪═══════════╪═══════╪═════╪════════╪═════╡
    │ 1.2   ┆ a         ┆ true  ┆ 4   ┆ -0.7   ┆ 2.5 │
    │ 5.077 ┆ transform ┆ false ┆ 4   ┆ 14.077 ┆ 4.0 │
    │ -2.29 ┆ pipedag   ┆ null  ┆ 2   ┆ 4.71   ┆ 0.0 │
    │ -0.0  ┆ cdegh     ┆ null  ┆ 0   ┆ 4.71   ┆ 0.0 │
    │ 3.0   ┆   -ade    ┆ true  ┆ 1   ┆ -0.7   ┆ 2.5 │
    │ -7.7  ┆   pq      ┆ true  ┆ 0   ┆ -0.7   ┆ 2.5 │
    └───────┴───────────┴───────┴─────┴────────┴─────┘
    """
    new = copy.copy(table)
    new._ast = Ungroup(table._ast)

    return new


@overload
def summarize(**kwargs: ColExpr) -> Pipeable: ...


@verb
@modify_ast
def summarize(table: Table, **kwargs: ColExpr) -> Pipeable:
    """
    Computes aggregates over groups of rows.

    :param kwargs:
        Each key is the name of a new column, and its value is the column
        expression defining the new column. The column expression may not contain
        columns that are neither part of the grouping columns nor wrapped in an
        aggregation function.


    In contrast to :doc:`pydiverse.transform.mutate`, this verb in general reduces the
    number of rows and only keeps the grouping columns and the new columns defined in
    the kwargs. One row for each unique combination of values in the grouping columns
    is created.


    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "a": [1.2, 5.077, -2.29, -0.0, 3.0, -7.7],
    ...         "b": ["a  ", "transform", "pipedag", "cdegh", "  -ade ", "  pq"],
    ...         "c": [True, False, None, None, True, True],
    ...     }
    ... )
    >>> (
    ...     t
    ...     >> group_by(t.c)
    ...     >> summarize(
    ...         u=t.b.str.len().mean(),
    ...         v=t.a.sum(filter=t.a >= 0),
    ...     )
    ...     >> show()
    ... )
    shape: (3, 3)
    ┌───────┬──────────┬───────┐
    │ c     ┆ u        ┆ v     │
    │ ---   ┆ ---      ┆ ---   │
    │ bool  ┆ f64      ┆ f64   │
    ╞═══════╪══════════╪═══════╡
    │ true  ┆ 4.666667 ┆ 4.2   │
    │ null  ┆ 6.0      ┆ -0.0  │
    │ false ┆ 9.0      ┆ 5.077 │
    └───────┴──────────┴───────┘
    """
    names, values = list(kwargs.keys()), list(kwargs.values())
    uuids = [uuid.uuid1() for _ in names]
    new = copy.copy(table)

    preprocessed = []
    for name, val in zip(names, values, strict=True):
        try:
            preprocessed.append(preprocess_arg(val, table, agg_is_window=False))
        except ErrorWithSource as e:
            raise type(e)(
                e.args[0] + f"\noccurred in `summarize` argument `{name}`\n"
                f"AST path to error source: {get_ast_path_str(val, e.source)}"
            ) from e
    new._ast = Summarize(table._ast, names, preprocessed, uuids)

    partition_by = set(table._cache.partition_by)

    if len(kwargs) == 0 and len(partition_by) == 0:
        raise ValueError("summarize without preceding group_by needs at least one column to summarize")

    def check_summarize_col_expr(expr: ColExpr, agg_fn_above: bool):
        if isinstance(expr, Col) and expr._uuid not in partition_by and not agg_fn_above:
            raise FunctionTypeError(
                f"column `{expr.ast_repr()}` is neither aggregated nor part of the grouping columns"
            )

        elif isinstance(expr, ColFn):
            if expr.op.ftype == Ftype.WINDOW:
                raise FunctionTypeError(f"forbidden window function `{expr.op.name}` in `summarize`")
            elif expr.op.ftype == Ftype.AGGREGATE and "partition_by" not in expr.context_kwargs:
                agg_fn_above = True

        for child in expr.iter_children():
            check_summarize_col_expr(child, agg_fn_above)

    for root in new._ast.values:
        check_summarize_col_expr(root, False)

    return new


@overload
def slice_head(n: int, *, offset: int = 0) -> Pipeable: ...


@verb
@modify_ast
def slice_head(table: Table, n: int, *, offset: int = 0) -> Pipeable:
    """
    Selects a subset of rows based on their index.

    :param n:
        The number of rows to select.

    :param offset:
        The index of the first row (0-based) that is included in the selection.


    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "a": [65, 5, 312, -55, 0],
    ...         "b": ["l", "r", "srq", "---", " "],
    ...     }
    ... )
    >>> t >> slice_head(3, offset=1) >> show()
    shape: (3, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    │ --- ┆ --- │
    │ i64 ┆ str │
    ╞═════╪═════╡
    │ 5   ┆ r   │
    │ 312 ┆ srq │
    │ -55 ┆ --- │
    └─────┴─────┘
    """
    errors.check_arg_type(int, "slice_head", "n", n)
    errors.check_arg_type(int, "slice_head", "offset", offset)

    if table._cache.partition_by:
        raise ValueError("cannot apply `slice_head` to a grouped table")

    new = copy.copy(table)
    new._ast = SliceHead(table._ast, n, offset)

    return new


@overload
def join(
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    how: Literal["inner", "left", "full"],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable: ...


@verb
def join(
    left: Table,
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    how: Literal["inner", "left", "full"],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable:
    """
    Joins two tables on a boolean expression.

    The left table in the join comes through the pipe `>>` operator from the
    left.

    :param right:
        The right table to join with.

    :param on:
        The join condition. If this is a list, the elements are joined via AND. Strings
        are interpreted as an equality condition on the columns with that name. See the
        note below for more information on which expressions are allowed.

    :param how:
        The join type.

    :param validate:
        Only relevant for polars. When set to ``"m:m"``, this does nothing. If set to
        ``"1:m"``, it is checked whether each right row matches at most one left row. In
        case this does not hold, an error is raised. Symmetrically, if set to ``"m:1"``
        it is checked whether each left row matches at most one right row. If set to
        ``"1:1"`` both ``"1:m"`` and ``"m:1"`` are checked.

    :param suffix:
        A string that is appended to all column names from the right table. If no suffix
        is specified and there are no column name collisions, columns will retain their
        original name. If there are name collisions, the name of the right table is
        appended to all columns of the right table. If this still does not resolve all
        name collisions, additionally an integer is appended to the column names of the
        right table. If the join condition only contains equalities of columns of the
        same name, only the columns of the right table colliding with a column of the
        left table are renamed.

    Note
    ----
    Not all backends can handle arbitrary boolean expressions in ``on`` with every join
    type.

    :polars:
        For everything except conjunctions of equalities, it depends on whether polars
        ``join_asof`` can handle the join condition.
    :postgres:
        For full joins, the join condition must be hashable or mergeable. See the
        `postgres documentation <https://wiki.postgresql.org/wiki/PostgreSQL_vs_SQL_Standard#FULL_OUTER_JOIN_conditions>`_
        for more details.

    Tip
    ---
    Two tables cannot be joined if one is derived from the other. In particular, before
    a self-join, the ``alias`` verb has to be applied to one table.

    Examples
    --------
    >>> t1 = pdt.Table({"a": [3, 1, 4, 1, 5, 9, 4]}, name="t1")
    >>> t2 = pdt.Table({"a": [4, 4, 1, 7], "b": ["f", "g", "h", "i"]}, name="t2")
    >>> t1 >> join(t2, t1.a == t2.a, how="left") >> show()
    shape: (9, 3)
    ┌─────┬──────┬──────┐
    │ a   ┆ a_t2 ┆ b_t2 │
    │ --- ┆ ---  ┆ ---  │
    │ i64 ┆ i64  ┆ str  │
    ╞═════╪══════╪══════╡
    │ 3   ┆ null ┆ null │
    │ 1   ┆ 1    ┆ h    │
    │ 4   ┆ 4    ┆ f    │
    │ 4   ┆ 4    ┆ g    │
    │ 1   ┆ 1    ┆ h    │
    │ 5   ┆ null ┆ null │
    │ 9   ┆ null ┆ null │
    │ 4   ┆ 4    ┆ f    │
    │ 4   ┆ 4    ┆ g    │
    └─────┴──────┴──────┘
    """

    errors.check_arg_type(Table, "join", "right", right)
    errors.check_arg_type(ColExpr | list | str, "join", "on", on)
    errors.check_arg_type(str | None, "join", "suffix", suffix)
    errors.check_literal_type(["inner", "left", "full"], "join", "how", how)
    errors.check_literal_type(["1:1", "1:m", "m:1", "m:m"], "join", "validate", validate)

    if left._cache.backend != right._cache.backend:
        raise TypeError("cannot join two tables with different backends")

    if left._cache.partition_by:
        raise ValueError(f"cannot join grouped table `{left._ast.short_name()}`")
    elif right._cache.partition_by:
        raise ValueError(f"cannot join grouped table `{right._ast.short_name()}`")

    if intersection := left._cache.derived_from & right._cache.derived_from:
        raise ValueError(
            f"table `{list(intersection)[0].name}` occurs twice in the table "
            "tree.\nhint: To join two tables derived from a common table, "
            "apply `>> alias()` to one of them before the join."
        )

    user_suffix = suffix
    if suffix is None and right._ast.name is not None:
        suffix = f"_{right._ast.name}"
    if suffix is None:
        suffix = "_right"

    left_names = set(left._cache.uuid_to_name[col._uuid] for col in left)
    right_names = set(right._cache.uuid_to_name[col._uuid] for col in right)

    # Lambda column resolution and checks for existence of columns are done manually
    # here since we need to incorporate columns from the right.
    def _preprocess_on(expr: ColExpr):
        if isinstance(expr, ColName):
            if expr in left:
                if expr in right:
                    raise ValueError(
                        f"`{expr.ast_repr()}` is ambiguous here. Please use the source "
                        "table of the column to reference it."
                    )
                return left[expr.name]
            if expr not in right:
                raise ValueError(f"no column with name `{expr.name}` found in the left or right table")
            return right[expr.name]

        if isinstance(expr, Col) and expr._uuid not in left._cache.cols and expr._uuid not in right._cache.cols:
            raise ValueError(
                f"column `{expr.ast_repr()}` used in `on` neither exists in the table "
                f"`{left._ast.short_name()}` nor in the table "
                f"`{right._ast.short_name()}`. The source table "
                f"`{expr._ast.short_name()}` of the column must be an "
                "ancestor of one of the two input tables."
            )

        return expr

    # bring join condition in standard expression form
    if not isinstance(on, list):
        on = [on]
    on = [left[expr] == right[expr] if isinstance(expr, str) else expr for expr in on]

    on = [pred.map_subtree(_preprocess_on) for pred in on]
    for i, pred in enumerate(on, 1):
        try:
            dtype = pred.dtype()
            pred.ftype(agg_is_window=False)
        except ErrorWithSource as e:
            raise type(e)(
                e.args[0] + f"\noccurred in {i}-{ordinal_suffix(i)} element of `on`\n"
                f"AST path to error source: {get_ast_path_str(pred, e.source)}"
            ) from e

        if types.without_const(dtype) != Bool():
            raise DataTypeError(
                f"{i}-{ordinal_suffix(i)} element of `on` has type `{pred.dtype()}`, "
                "but join conditions must have boolean type"
            )

    # apply `rename` to the right table if necessary
    if user_suffix:
        for name in right._cache.name_to_uuid.keys():
            if name + suffix in left_names:
                raise ValueError(
                    f"column name `{name + suffix}` appears both in the left and right "
                    f"table using the user-provided suffix `{suffix}`\n"
                    "hint: Specify a different suffix to prevent name collisions or "
                    "no suffix for automatic name collision resolution."
                )

        # If the user actively specifies a suffix, we apply it to all right columns.
        right >>= rename({col: col.name + user_suffix for col in right})

    elif right_names & left_names:
        cnt = 0
        for name in right_names:
            suffixed = name + suffix + (f"_{cnt}" if cnt > 0 else "")
            while suffixed in left_names:
                cnt += 1
                suffixed = name + suffix + f"_{cnt}"

        if cnt > 0:
            suffix += f"_{cnt}"

        on_uuids = set(
            col._uuid for col in itertools.chain(*(pred.iter_subtree_preorder() for pred in on)) if isinstance(col, Col)
        )
        right_on_names = set(col.name for col in right if col._uuid in on_uuids)

        if not (right_names - right_on_names) & left_names:
            # If nothing except join columns clashes, we only rename the clashing
            # columns on the right.
            right >>= rename({col: col.name + suffix for col in right if col.name in left_names})

        else:
            right >>= rename({col: col.name + suffix for col in right})

    if len(on) == 0:
        on = LiteralCol(True)
    else:
        on = functools.reduce(operator.and_, on[1:], on[0])

    if how == "full" and not all(pred.op == ops.equal for pred in split_join_cond(on)):
        raise ValueError("in a `full` join, only equality predicates can be used")

    for fn in on.iter_subtree_postorder():
        if isinstance(fn, ColFn) and fn.op.ftype != Ftype.ELEMENT_WISE:
            raise FunctionTypeError(
                f"window function `{fn.op.name}` not allowed in `on`\n"
                "hint: First add the result of the window function to the table using "
                "`mutate`."
            )

    new = copy.copy(left)
    new._ast = Join(left._ast, right._ast, on, how, validate)

    new, left = check_subquery(new, left)
    new, right = check_subquery(new, right, is_right=True)

    new._cache = left._cache.update(new._ast, right_cache=right._cache)

    return new


# We define the join variations explicitly instead of via functools.partial since vscode
# gives functools.partial objects a different color than normal python functions which
# looks very confusing.
@overload
def inner_join(
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable: ...


@verb
def inner_join(
    left: Table,
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable:
    """
    Alias for the :doc:`pydiverse.transform.join` verb with ``how="inner"``.
    """

    return left >> join(right, on, "inner", validate=validate, suffix=suffix)


@overload
def left_join(
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable: ...


@verb
def left_join(
    left: Table,
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable:
    """
    Alias for the :doc:`pydiverse.transform.join` verb with ``how="left"``.
    """

    return left >> join(right, on, "left", validate=validate, suffix=suffix)


@overload
def full_join(
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable: ...


@verb
def full_join(
    left: Table,
    right: Table,
    on: ColExpr[Bool] | str | list[ColExpr[Bool] | str],
    *,
    validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    suffix: str | None = None,
) -> Pipeable:
    """
    Alias for the :doc:`pydiverse.transform.join` verb with ``how="full"``.
    """

    return left >> join(right, on, "full", validate=validate, suffix=suffix)


@overload
def cross_join(
    right: Table,
    *,
    suffix: str | None = None,
) -> Pipeable: ...


@verb
def cross_join(
    left: Table,
    right: Table,
    *,
    suffix: str | None = None,
) -> Pipeable:
    """
    Alias for the :doc:`pydiverse.transform.join` verb with an empty ``on`` clause.
    """

    return left >> join(right, how="inner", on=[], suffix=suffix)


@overload
def union(
    right: Table,
    *,
    distinct: bool = False,
) -> Pipeable: ...


@overload
def union(
    left: Table,
    right: Table,
    *,
    distinct: bool = False,
) -> Table: ...


def _union_impl(
    left: Table,
    right: Table,
    *,
    distinct: bool = False,
) -> Table:
    """
    Unions two tables by stacking rows vertically.

    The left table in the union comes through the pipe `>>` operator from the
    left.

    :param right:
        The right table to union with.

    :param distinct:
        If ``True``, performs UNION (removes duplicates). If ``False``,
        performs UNION ALL (keeps duplicates).

    Note
    ----
    Both tables must have the same number of columns with compatible types.
    Column names must match between the two tables.

    Examples
    --------
    >>> t1 = pdt.Table({"a": [1, 2, 3], "b": [4, 5, 6]}, name="t1")
    >>> t2 = pdt.Table({"a": [7, 8], "b": [9, 10]}, name="t2")
    >>> t1 >> union(t2) >> show()
    shape: (5, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    │ --- ┆ --- │
    │ i64 ┆ i64 │
    ╞═════╪═════╡
    │ 1   ┆ 4   │
    │ 2   ┆ 5   │
    │ 3   ┆ 6   │
    │ 7   ┆ 9   │
    │ 8   ┆ 10  │
    └─────┴─────┘
    """
    errors.check_arg_type(Table, "union", "right", right)

    if left._cache.backend != right._cache.backend:
        raise TypeError("cannot union two tables with different backends")

    if left._cache.partition_by:
        raise ValueError(f"cannot union grouped table `{left._ast.short_name()}`")
    elif right._cache.partition_by:
        raise ValueError(f"cannot union grouped table `{right._ast.short_name()}`")

    # Check that both tables have the same columns
    left_cols = set(left._cache.name_to_uuid.keys())
    right_cols = set(right._cache.name_to_uuid.keys())

    if left_cols != right_cols:
        missing_left = right_cols - left_cols
        missing_right = left_cols - right_cols
        error_msg = "tables must have the same columns for union"
        if missing_left:
            error_msg += f"\n  columns in right but not in left: {sorted(missing_left)}"
        if missing_right:
            error_msg += f"\n  columns in left but not in right: {sorted(missing_right)}"
        raise ValueError(error_msg)

    # Check column type compatibility using lca_type
    from pydiverse.transform._internal.tree.types import lca_type

    for col_name in left_cols:
        left_col = left._cache.cols[left._cache.name_to_uuid[col_name]]
        right_col = right._cache.cols[right._cache.name_to_uuid[col_name]]
        left_dtype = left_col.dtype()
        right_dtype = right_col.dtype()

        # Check if types are compatible by trying to find a common ancestor
        try:
            lca_type([left_dtype, right_dtype])
        except DataTypeError as e:
            raise TypeError(
                f"column '{col_name}' has incompatible types: left has {left_dtype}, right has {right_dtype}"
            ) from e

    new = copy.copy(left)
    new._ast = Union(left._ast, right._ast, distinct)

    new, left = check_subquery(new, left)
    new, right = check_subquery(new, right, is_right=True)

    new._cache = left._cache.update(new._ast, right_cache=right._cache)

    return new


def union(
    left_or_right: Table,
    right: Table | None = None,
    *,
    distinct: bool = False,
) -> Table | Pipeable:
    """
    Unions two tables by stacking rows vertically.

    The left table in the union comes through the pipe `>>` operator from the
    left.

    :param right:
        The right table to union with.

    :param distinct:
        If ``True``, performs UNION (removes duplicates). If ``False``,
        performs UNION ALL (keeps duplicates).

    Note
    ----
    Both tables must have the same number of columns with compatible types.
    Column names must match between the two tables.

    Examples
    --------
    >>> t1 = pdt.Table({"a": [1, 2, 3], "b": [4, 5, 6]}, name="t1")
    >>> t2 = pdt.Table({"a": [7, 8], "b": [9, 10]}, name="t2")
    >>> t1 >> union(t2) >> show()
    shape: (5, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    │ --- ┆ --- │
    │ i64 ┆ i64 │
    ╞═════╪═════╡
    │ 1   ┆ 4   │
    │ 2   ┆ 5   │
    │ 3   ┆ 6   │
    │ 7   ┆ 9   │
    │ 8   ┆ 10  │
    └─────┴─────┘

    You can also call it directly:
    >>> union(t1, t2) >> show()
    """
    # If called with two arguments directly (union(tbl1, tbl2)), return Table directly
    if right is not None:
        return _union_impl(left_or_right, right, distinct=distinct)  # returns Table

    # If called with one argument (union(tbl2)), behave like a verb for pipe syntax
    # Create a verb wrapper that will be called by the pipe operator
    # Note: left_or_right is the right table in this branch (since right is None)
    @verb
    def _union_verb(left_table: Table, right_table: Table, distinct: bool) -> Table:
        return _union_impl(left_table, right_table, distinct=distinct)

    return _union_verb(left_or_right, distinct=distinct)  # returns Pipeable


@overload
def show(pipe: bool = False) -> Pipeable | None: ...


@verb
def show(table: Table, pipe: bool = False) -> Pipeable | None:
    """
    Prints the table to stdout.

    :param pipe:
        If set to `True`, the table is returned, else `None` is returned.

    Note
    ----
    During interactive development in the python shell, you usually want to keep
    `pipe = False`, else the table is printed twice. The main use for `pipe = True` is
    when you don't want to break a long sequence of verbs in a file.
    """
    print(table)
    return table if pipe else None


@overload
def name() -> str | None: ...


@verb
def name(table: Table) -> str | None:
    """
    Returns the name of the table.
    """
    return table._ast.name


@overload
def columns() -> list[str]: ...


@verb
def columns(table: Table) -> list[str]:
    """
    Returns the column names of the table.
    """
    return list(table._cache.name_to_uuid.keys())


@overload
def ast_repr(verb_depth: int = 7, expr_depth: int = 2, pipe: bool = False) -> Pipeable | None: ...


@verb
def ast_repr(table: Table, verb_depth: int = 7, expr_depth: int = 2, *, pipe: bool = False) -> Pipeable | None:
    r"""
    Prints the AST of the table to stdout.

    :param pipe:
        If set to `True`, the table is returned, else `None` is returned.

    Note
    ----
    During interactive development in the python shell, you usually want to keep
    `pipe = False`, else the table is printed, too. The main use for `pipe = True` is
    when you don't want to break a long sequence of verbs in a file.

    Examples
    --------
    >>> tbl1 = Table(dict(a=[1, 2], b=["x", "y"]), name="tbl1")
    >>> tbl2 = Table(dict(a=[1, 2], c=["z", "zz"]), name="tbl2")
    >>> (
    ...     tbl1
    ...     >> mutate(u=C.a + 5)
    ...     >> select(tbl1.a)
    ...     >> left_join(
    ...         tbl2
    ...         >> arrange(tbl2.a)
    ...         >> filter(tbl2.c.str.len() <= 10)
    ...         >> full_join(s := tbl2 >> alias("s"), on="c"),
    ...         on="a",
    ...     )
    ...     >> slice_head(32)
    ...     >> ast_repr()
    ... )
    tbl2 = Table(df, Polars(), name="tbl2")
    tbl1 = Table(df, Polars(), name="tbl1")

    (
        tbl1
        >> mutate(u=tbl1.a + 5)
        >> select(tbl1.a)
        >> join(
            tbl2
            >> arrange(tbl2.a)
            >> filter(tbl2.c.str.len() <= 10)
            >> join(
                tbl2 >> alias("s") >> rename({"a": "a_s", "c": "c_s"}),
                how="full",
                on=tbl2.c == s.c,
                validate="m:m",
            )
            >> rename({"a": "a_tbl2"}),
            how="left",
            on=tbl1.a == tbl2.a,
            validate="m:m",
        )
        >> slice_head(n=32, offset=0)
    )
    """

    print(table._ast.ast_repr(verb_depth, expr_depth), end="")
    return table if pipe else None


def preprocess_arg(arg: ColExpr, table: Table, *, agg_is_window: bool = True) -> Any:
    arg = wrap_literals(arg)
    assert isinstance(arg, ColExpr | Order)

    def _preprocess_expr(expr: ColExpr, eval_aligned: bool = False):
        if isinstance(expr, Col) and expr._uuid not in table._cache.cols and not eval_aligned:
            raise ColumnNotFoundError(f"column `{expr.ast_repr()}` does not exist in table `{table._ast.name}`")

        if not eval_aligned and isinstance(expr, Series):
            raise TypeError(
                f"invalid use of series `{expr.ast_repr()}` in a column expression"
                "\nnote: polars / pandas series must be wrapped in `eval_aligned` for "
                "use in pydiverse.transform"
            )

        if (
            agg_is_window
            and isinstance(expr, ColFn)
            and "partition_by" not in expr.context_kwargs
            and (expr.op.ftype in (Ftype.WINDOW, Ftype.AGGREGATE))
        ):
            expr.context_kwargs["partition_by"] = [table._cache.cols[uid] for uid in table._cache.partition_by]

        if isinstance(expr, ColName):
            return table[expr.name]

        new = copy.copy(expr)
        new.map_children(
            functools.partial(
                _preprocess_expr,
                eval_aligned=eval_aligned | isinstance(expr, EvalAligned),
            )
        )

        # add casts for boolean add / sum
        # If we have more operations like these, which we want to map to other
        # operations on the AST, consider streamlining this in some special function
        if (
            isinstance(new, ColFn)
            and len(new.args) > 0
            and types.without_const(new.args[0].dtype()) == Bool()
            and new.op in (ops.add, ops.sum)
        ):
            new.args = [arg.cast(Int64) for arg in new.args]

        return new

    res = _preprocess_expr(arg)
    assert res.dtype() is not None
    assert res.ftype(agg_is_window=agg_is_window) is not None
    return res


def ordinal_suffix(n: int):
    return {1: "st", 2: "nd"}.get(n % 10, "th")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from datetime import datetime

import pydiverse.transform as pdt
from pydiverse.transform._internal.errors import DataTypeError
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_eq(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 == datetime(1970, 1, 1)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 == datetime(2004, 12, 31)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 == C.col2))


def test_nq(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 != datetime(1970, 1, 1)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 != datetime(2004, 12, 31)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 != C.col2))


def test_lt(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 < datetime(1970, 1, 1)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 < datetime(2004, 12, 31)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 < C.col2))


def test_gt(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 > datetime(1970, 1, 1)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 > datetime(2004, 12, 31)))
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 > C.col2))


def test_le(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 <= C.col2))


def test_ge(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 >= C.col2))


def test_year(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.year(),
            y=C.col2.dt.year(),
            z=t.cdate.dt.year(),
        ),
    )


def test_month(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.month(),
            y=C.col2.dt.month(),
            z=t.cdate.dt.month(),
        ),
    )


def test_day(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t >> mutate(x=C.col1.dt.day(), y=C.col2.dt.day(), z=t.cdate.dt.day()),
    )


def test_hour(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.hour(),
            y=C.col2.dt.hour(),
        ),
    )

    assert_result_equal(
        df_datetime,
        lambda t: t >> mutate(z=t.cdate.dt.hour()),
        exception=DataTypeError,
    )


def test_minute(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.minute(),
            y=C.col2.dt.minute(),
        ),
    )


def test_second(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.second(),
            y=C.col2.dt.second(),
        ),
    )


def test_millisecond(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.millisecond(),
            y=C.col2.dt.millisecond(),
        ),
    )


def test_microseconds(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.microsecond(),
            y=C.col2.dt.microsecond(),
        ),
    )


def test_day_of_week(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.day_of_week(),
            y=C.col2.dt.day_of_week(),
        ),
    )


def test_day_of_year(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.day_of_year(),
            y=C.col2.dt.day_of_year(),
        ),
    )


def test_cast(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=t.cdate.cast(pdt.Datetime()),
            y=t.col1.cast(pdt.Date()),
        ),
    )


# def test_duration_add(df_datetime):
#     assert_result_equal(df_datetime, lambda t: t >> mutate(z=t.cdur + t.cdur))


# def test_dt_subtract(df_datetime):
#     assert_result_equal(df_datetime, lambda t: t >> mutate(z=t.col1 - t.col2))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform._internal.errors import FunctionTypeError, SubqueryError
from pydiverse.transform.extended import *
from tests.util import assert_result_equal


def test_simple_ungrouped(df3):
    assert_result_equal(
        df3,
        lambda t: t >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )


def test_simple_grouped(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1) >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )


def test_partition_by_argument(df3, df4):
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(
            u=t.col1.min(partition_by=t.col3),
            v=t.col4.sum(partition_by=t.col2),
            w=pdt.rank(
                arrange=[t.col5.descending().nulls_last(), t.col4.nulls_first()],
                partition_by=[t.col2],
            ),
            x=pdt.row_number(arrange=[t.col4.nulls_last()], partition_by=[t.col1, t.col2]),
        ),
    )

    assert_result_equal(
        (df3, df4),
        lambda t, u: t
        >> join(u, t.col1 == u.col3, how="left")
        >> group_by(t.col2)
        >> mutate(y=(u.col3 + t.col1).max(partition_by=(col for col in t if col.name != "col7"))),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            u=t.col3.sum(),
            v=t.col2.sum(partition_by=[t.col2]),
        ),
    )


def test_chained(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1) >> mutate(min=t.col4.min()) >> mutate(max=t.col4.max(), mean=t.col4.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1) >> mutate(min=t.col4.min(), max=t.col4.max()) >> mutate(span=C.max - C.min),
    )


def test_nested(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(range=t.col4.max() - 10)
        >> ungroup()
        >> alias()
        >> mutate(range_mean=C.range.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=C.col4.max()) >> mutate(y=C.x.min() * 1) >> mutate(z=C.y.mean()) >> mutate(w=C.x / C.y),
        may_throw=True,
        exception=SubqueryError,
    )

    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=(C.col4.max().min() + C.col2.mean()).max()),
        exception=FunctionTypeError,
        may_throw=True,
    )


def test_filter(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> mutate(mean3=t.col3.mean()) >> alias() >> filter(C.mean3 <= 2.0),
    )


def test_filter_argument(df3, df4):
    assert_result_equal(df4, lambda t: t >> mutate(u=t.col2.mean(filter=~t.col2.is_null())))

    assert_result_equal(df4, lambda t: t >> mutate(u=t.col2.mean(filter=~(t.col4 % 3 == 0))))

    assert_result_equal(df3, lambda t: t >> mutate(u=t.col4.sum(partition_by=t.col2)))

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=t.col1.min(filter=(~t.col1.is_null()), partition_by=[t.col3]),
            v=t.col4.max(filter=~t.col4.is_null(), partition_by=[t.col1]),
        ),
    )

    assert_result_equal(df4, lambda t: t >> mutate(u=t.col3.min(filter=t.col3.is_null())))


def test_arrange(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> mutate(mean3=t.col3.mean()) >> arrange(C.mean3),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(-t.col4) >> group_by(t.col1, t.col2) >> mutate(mean3=t.col3.mean()) >> arrange(C.mean3),
    )


def test_summarize(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(range=t.col4.max() - t.col4.min())
        >> alias()
        >> summarize(mean_range=C.range.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarize(range=t.col4.max() - t.col4.min())
        >> alias()
        >> mutate(mean_range=C.range.mean()),
    )


def test_intermediate_select(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(x=t.col4.mean())
        >> alias()
        >> mutate(y=C.x.min())
        # TODO: technically, we could remove some window functions here and prevent
        # subqueries
        >> alias()
        >> mutate(z=(C.x - C.y).mean())
        >> select(C.z),
    )


def test_arrange_argument(df3):
    # Grouped
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1) >> mutate(x=C.col4.shift(1, arrange=C.col3.nulls_last())) >> select(C.x),
    )

    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col2) >> mutate(x=pdt.row_number(arrange=C.col4.descending())) >> select(C.x),
    )

    # Ungrouped
    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=C.col4.shift(1, arrange=[-C.col3])) >> select(C.x),
    )

    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=pdt.row_number(arrange=[-C.col4])) >> select(C.x),
    )


def test_complex(df3):
    # Window function before summarize
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(mean3=t.col3.mean(), rn=pdt.row_number(arrange=[C.col1, C.col2]))
        >> alias()
        >> filter(C.mean3 > C.rn)
        >> alias()
        >> summarize(meta_mean=C.mean3.mean())
        >> filter(C.col1 >= C.meta_mean)
        >> filter(C.col1 != 1)
        >> arrange(C.meta_mean),
    )

    # Window function after summarize
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarize(mean3=t.col3.mean())
        >> alias()
        >> mutate(minM3=C.mean3.min(), maxM3=C.mean3.max())
        >> mutate(span=C.maxM3 - C.minM3)
        >> alias()
        >> filter(C.span < 3)
        >> arrange(C.span),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarize(mean3=t.col3.mean(), u=t.col4.max())
        >> group_by(C.u)
        >> alias()
        >> mutate(minM3=C.mean3.min(), maxM3=C.mean3.max())
        >> mutate(span=C.maxM3 - C.minM3)
        >> alias()
        >> filter(C.span < 3)
        >> arrange(C.span),
    )


def test_nested_bool(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(x=t.col1 <= t.col2, y=(t.col3 * 4) >= C.col4)
        >> mutate(
            xshift=C.x.shift(1, arrange=[t.col4.nulls_last()]),
            yshift=C.y.shift(-1, arrange=[t.col4.nulls_first()]),
        )
        >> mutate(xAndY=C.x & C.y, xAndYshifted=C.xshift & C.yshift),
    )


# Test specific operations


def test_op_shift(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            shift1=t.col2.shift(1, arrange=[t.col4.nulls_first()]),
            shift2=t.col4.shift(-2, 0, arrange=[t.col4.nulls_last()]),
            shift3=t.col4.shift(0, arrange=[t.col4.nulls_first()]),
            u=C.col1.shift(1, 0, arrange=[t.col4.nulls_last()]),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=t.col1.shift(1, 0, arrange=[t.col2.nulls_last(), t.col4.nulls_first()]),
            v=t.col1.shift(2, 1, arrange=[t.col4.descending().nulls_first()]),
        ),
    )


def test_op_row_number(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            row_number1=pdt.row_number(arrange=[C.col4.descending().nulls_last()]),
            row_number2=pdt.row_number(arrange=[C.col2.nulls_last(), C.col3.nulls_first(), t.col4.nulls_last()]),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=pdt.row_number(arrange=[C.col4.descending().nulls_last()]),
            v=pdt.row_number(arrange=[t.col3.descending().nulls_first(), t.col4.nulls_first()]),
        ),
    )


def test_op_rank(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            rank1=pdt.rank(arrange=[t.col1.nulls_last()]),
            rank2=pdt.rank(arrange=[t.col2.nulls_first()]),
            rank3=pdt.rank(arrange=[t.col2.nulls_last()]),
            rank4=pdt.rank(arrange=[t.col5.nulls_first()]),
            rank5=pdt.rank(arrange=[t.col5.descending().nulls_first()]),
            rank_expr=pdt.rank(arrange=[(t.col3 - t.col2).nulls_last()]),
        ),
    )


def test_op_dense_rank(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            rank1=pdt.dense_rank(arrange=[t.col5.nulls_first()]),
            rank2=pdt.dense_rank(arrange=[t.col2.nulls_last()]),
            rank3=pdt.dense_rank(arrange=[t.col2.nulls_last()]),
        )
        >> ungroup()
        >> mutate(
            rank4=pdt.dense_rank(arrange=[t.col4.nulls_first()], partition_by=[t.col2]),
            rank5=pdt.dense_rank(
                arrange=[t.col5.descending().nulls_first()],
                partition_by=[t.col2],
            ),
        ),
    )


def test_partition_by_const_col(df3):
    assert_result_equal(df3, lambda t: t >> mutate(x=0) >> mutate(y=t.col3.sum(partition_by=C.x)))


def test_cum_sum(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            p=t.col1.cum_sum(arrange=t.col4.nulls_last()),
            r=t.col4.cum_sum(arrange=t.col4.nulls_first()),
        ),
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import ContextKwarg
from pydiverse.transform._internal.ops.ops.aggregation import Aggregation
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import List, S

list_agg = Aggregation(
    "list.agg",
    Signature(S, return_type=List(S)),
    context_kwargs=[
        ContextKwarg("partition_by"),
        ContextKwarg("arrange"),
        ContextKwarg("filter"),
    ],
    doc="""
Collect the elements of each group in a list.
""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform import C
from pydiverse.transform._internal.pipe.verbs import (
    arrange,
    mutate,
)
from tests.util import assert_result_equal


def test_noop(df1):
    assert_result_equal(df1, lambda t: t >> arrange(), may_throw=True, exception=TypeError)


def test_arrange(df2):
    assert_result_equal(df2, lambda t: t >> arrange(t.col1.ascending()))
    assert_result_equal(df2, lambda t: t >> arrange((-t.col1).descending()))
    assert_result_equal(df2, lambda t: t >> arrange(t.col3))
    assert_result_equal(df2, lambda t: t >> arrange(-t.col3))


def test_arrange_expression(df3):
    assert_result_equal(df3, lambda t: t >> arrange(t.col2, t.col4), check_row_order=True)
    assert_result_equal(df3, lambda t: t >> arrange(-t.col4 * 2), check_row_order=True)


def test_arrange_null(df2):
    assert_result_equal(df2, lambda t: t >> arrange(t.col2))
    assert_result_equal(df2, lambda t: t >> arrange(-t.col2))


def test_multiple(df3):
    assert_result_equal(df3, lambda t: t >> arrange(t.col2, -t.col3, -t.col4))

    assert_result_equal(df3, lambda t: t >> arrange(t.col2) >> arrange(-t.col3, -t.col4))

    assert_result_equal(
        df3,
        lambda t: t >> arrange(t.col2, -t.col2),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(t.col2, C.col2),
    )


def test_nulls_first(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> arrange(
            t.col1.nulls_first(),
            t.col2.descending().nulls_first(),
            t.col5.nulls_first(),
        ),
        check_row_order=True,
    )


def test_nulls_last(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> arrange(
            t.col1.nulls_last(),
            t.col2.nulls_last().descending(),
            t.col5.nulls_last(),
        ),
        check_row_order=True,
    )


def test_nulls_first_last_mixed(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> arrange(
            t.col1.nulls_first(),
            t.col2.nulls_last().descending(),
            t.col5.descending().nulls_last(),
        ),
        check_row_order=True,
    )


def test_arrange_after_mutate(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=t.col1 <= t.col2) >> arrange(C.x.nulls_last(), C.col4.nulls_first()),
        check_row_order=True,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: A004

from ._internal.pipe.functions import (
    all,
    any,
    coalesce,
    count,
    dense_rank,
    lit,
    max,
    min,
    rand,
    rank,
    row_number,
    sum,
    when,
)
from ._internal.pipe.verbs import filter
from .common import *  # noqa: F403
from .common import __all__ as __common

__all__ = __common + [
    "any",
    "all",
    "count",
    "sum",
    "filter",
    "coalesce",
    "dense_rank",
    "max",
    "rand",
    "min",
    "rank",
    "row_number",
    "when",
    "lit",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    NUMERIC,
    Bool,
    Const,
    Float,
    Int,
)

pow = Operator(
    "__pow__",
    Signature(Int(), Int(), return_type=Float()),
    Signature(Float(), Float(), return_type=Float()),
    doc="""
Computes the power x ** y.

Note
----
Polars throws on negative exponents in the integer case. A polars error like
`failed to convert X to u32` may be due to negative inputs to this function.
""",
)


neg = Operator(
    "__neg__",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="The unary - (negation) operator (__neg__)",
)

pos = Operator(
    "__pos__",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="The unary + operator (__pos__)",
)

abs = Operator(
    "abs",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="Computes the absolute value.",
)

round = Operator(
    "round",
    *(Signature(t, Const(Int()), return_type=t) for t in NUMERIC),
    param_names=["self", "decimals"],
    default_values=[..., 0],
    doc="""
Rounds to a given number of decimals.

:param decimals:
    The number of decimals to round by.
""",
)

floor = Operator(
    "floor",
    Signature(Float(), return_type=Float()),
    doc="Returns the largest integer less than or equal to the input.",
)

ceil = Operator(
    "ceil",
    Signature(Float(), return_type=Float()),
    doc="Returns the smallest integer greater than or equal to the input.",
)

log = Operator(
    "log",
    Signature(Float(), return_type=Float()),
    doc="Computes the natural logarithm.",
)

exp = Operator(
    "exp",
    Signature(Float(), return_type=Float()),
    doc="Computes the exponential function.",
)

log10 = Operator(
    "log10",
    Signature(Float(), return_type=Float()),
    doc="Computes the base-10 logarithm.",
)

sin = Operator("sin", Signature(Float(), return_type=Float()), doc="Computes the sine.")

cos = Operator("cos", Signature(Float(), return_type=Float()), doc="Computes the cosine.")

tan = Operator("tan", Signature(Float(), return_type=Float()), doc="Computes the tangent.")

asin = Operator(
    "asin",
    Signature(Float(), return_type=Float()),
    doc="Computes the inverse sine.",
)

acos = Operator(
    "acos",
    Signature(Float(), return_type=Float()),
    doc="Computes the inverse cosine.",
)

atan = Operator(
    "atan",
    Signature(Float(), return_type=Float()),
    doc="Computes the inverse tangent.",
)


sqrt = Operator("sqrt", Signature(Float(), return_type=Float()), doc="Computes the square root.")

cbrt = Operator("cbrt", Signature(Float(), return_type=Float()), doc="Computes the cube root.")


is_inf = Operator(
    "is_inf",
    Signature(Float(), return_type=Bool()),
    doc="""
Whether the number is infinite.

Note
----
This is currently only useful for backends supporting IEEE 754-floats. On
other backends it always returns False.
""",
)

is_not_inf = Operator("is_not_inf", Signature(Float(), return_type=Bool()))

is_nan = Operator("is_nan", Signature(Float(), return_type=Bool()))

is_not_nan = Operator("is_not_nan", Signature(Float(), return_type=Bool()))

rand = Operator(
    "rand",
    Signature(return_type=Float()),
    generate_expr_method=False,
    doc="Generates a column of random floating point number between 0 and 1.",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from ._internal.errors import (
    ColumnNotFoundError,
    DataTypeError,
    FunctionTypeError,
    NotSupportedError,
    SubqueryError,
)

__all__ = [
    "SubqueryError",
    "DataTypeError",
    "FunctionTypeError",
    "NotSupportedError",
    "ColumnNotFoundError",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.tree.col_expr import ColName


class MC(type):
    def __getattr__(cls, name: str) -> ColName:
        return ColName(name)

    def __getitem__(cls, name: str) -> ColName:
        return ColName(name)


class C(metaclass=MC):
    """As an alternative to referencing a column via `<table>.<column name>`, you can
    use `C.<column name>` or `C[<column name>]`. Using :class:`C` is necessary if the
    column to be referenced does not live in a table stored as a python variable."""

    pass



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.ops.op import ContextKwarg, Ftype, Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    COMPARABLE,
    NUMERIC,
    Bool,
    Const,
    Float,
    Int,
    S,
    String,
)


class Aggregation(Operator):
    def __init__(
        self,
        name: str,
        *signatures: Signature,
        context_kwargs=None,
        param_names=None,
        default_values=None,
        generate_expr_method: bool = True,
        doc: str = "",
    ):
        if context_kwargs is None:
            context_kwargs = [
                ContextKwarg("partition_by", False),
                ContextKwarg("filter", False),
            ]
        super().__init__(
            name,
            *signatures,
            ftype=Ftype.AGGREGATE,
            context_kwargs=context_kwargs,
            param_names=param_names,
            default_values=default_values,
            generate_expr_method=generate_expr_method,
            doc=doc,
        )


min = Aggregation(
    "min",
    *(Signature(dtype, return_type=dtype) for dtype in COMPARABLE),
    doc="Computes the minimum value in each group.",
)

max = Aggregation(
    "max",
    *(Signature(dtype, return_type=dtype) for dtype in COMPARABLE),
    doc="Computes the maximum value in each group.",
)

mean = Aggregation(
    "mean",
    Signature(Float(), return_type=Float()),
    Signature(Int(), return_type=Float()),
    doc="Computes the average value in each group.",
)

sum = Aggregation(
    "sum",
    *(Signature(dtype, return_type=dtype) for dtype in NUMERIC),
    Signature(Bool(), return_type=Int()),
    doc="Computes the sum of values in each group.",
)

any = Aggregation(
    "any",
    Signature(Bool(), return_type=Bool()),
    doc="Indicates whether at least one value in a group is True.",
)

all = Aggregation(
    "all",
    Signature(Bool(), return_type=Bool()),
    doc="Indicates whether every non-null value in a group is True.",
)

count = Aggregation(
    "count",
    Signature(S, return_type=Int()),
    doc="""
Counts the number of non-null elements in the column.
""",
)

count_star = Aggregation(
    "count",
    Signature(return_type=Int()),
    generate_expr_method=False,
    doc="""
Returns the number of rows of the current table, like :code:`COUNT(*)` in SQL.
""",
)

str_join = Aggregation(
    "str.join",
    Signature(String(), Const(String()), return_type=String()),
    param_names=["self", "delimiter"],
    default_values=[..., ""],
    context_kwargs=[
        ContextKwarg("partition_by"),
        ContextKwarg("filter"),
        ContextKwarg("arrange"),
    ],
    doc="""
Concatenates all strings in a group to a single string.

:param delimiter:
    The string to insert between the elements.""",
)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import sqlalchemy as sqa
from sqlalchemy.dialects.postgresql import aggregate_order_by

from pydiverse.common import Bool, Dtype, Float, Float32, Int, Int32, Int64, String
from pydiverse.transform._internal.backend.sql import SqlImpl
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.col_expr import Cast, ColFn


class PostgresImpl(SqlImpl):
    backend_name = "postgres"

    @classmethod
    def compile_cast(cls, cast: Cast, sqa_col: dict[str, sqa.Label]) -> Cast:
        if types.without_const(cast.val.dtype()).is_float():
            if cast.target_type.is_int():
                return cls.cast_compiled(cast, sqa.func.trunc(cls.compile_col_expr(cast.val, sqa_col)))

            if cast.target_type == String():
                compiled = super().compile_cast(cast, sqa_col)
                return sqa.case(
                    (compiled == "NaN", "nan"),
                    (compiled == "Infinity", "inf"),
                    (compiled == "-Infinity", "-inf"),
                    else_=compiled,
                )

        if types.without_const(cast.val.dtype()) == Bool() and cast.target_type == Int64():
            # postgres does not like casts bool -> bigint, so we go via int
            return cls.compile_cast(Cast(cast.val, Int32()).cast(Int64()), sqa_col)

        return super().compile_cast(cast, sqa_col)

    @classmethod
    def cast_compiled(cls, cast: Cast, compiled_expr: sqa.ColumnElement) -> sqa.Cast:
        if not cast.strict:
            cast.strict = True  # postgres does not have TRY_CAST

            if types.without_const(cast.val.dtype()) == String():
                compiled_expr = sqa.case(
                    (
                        sqa.func.pg_input_is_valid(
                            compiled_expr,
                            str(cls.sqa_type(cast.target_type)).lower().replace(" ", ""),
                        ),
                        compiled_expr,
                    ),
                    else_=sqa.null(),
                )

            def int_type_range(dtype: Int) -> tuple[int, int]:
                is_signed = dtype.__class__.__name__[0] == "I"
                bits = int(dtype.__class__.__name__[4 - is_signed :])

                if is_signed:
                    return (-(2 ** (bits - 1)), 2 ** (bits - 1) - 1)
                else:
                    return (0, 2**bits - 1)

            if cast.val.dtype().is_int() and cast.target_type.is_int():
                source_range = int_type_range(cast.val.dtype())
                target_range = int_type_range(cast.target_type)

                if source_range[0] < target_range[0]:
                    compiled_expr = sqa.case(
                        (compiled_expr >= target_range[0], compiled_expr),
                        else_=sqa.null(),
                    )
                if source_range[1] > target_range[1]:
                    compiled_expr = sqa.case(
                        (compiled_expr <= target_range[1], compiled_expr),
                        else_=sqa.null(),
                    )

            if cast.val.dtype().is_float() and cast.target_type.is_int():
                target_range = int_type_range(cast.target_type)

                compiled_expr = sqa.case(
                    (
                        (compiled_expr != cls.nan())
                        & (compiled_expr != cls.inf())
                        & (compiled_expr != -cls.inf())
                        & (compiled_expr >= target_range[0])
                        & (compiled_expr <= target_range[1]),
                        compiled_expr,
                    ),
                    else_=sqa.null(),
                )

        return super().cast_compiled(cast, compiled_expr)

    @classmethod
    def sqa_type(cls, pdt_type: Dtype):
        if isinstance(pdt_type, types.List):
            return sqa.types.ARRAY(item_type=cls.sqa_type())
        if isinstance(pdt_type, Float32):
            return sqa.REAL()

        return super().sqa_type(pdt_type)

    @classmethod
    def fix_fn_types(cls, fn: ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement) -> sqa.ColumnElement:
        if isinstance(fn.op, ops.DatetimeExtract | ops.DateExtract):
            return sqa.cast(val, sqa.BigInteger)
        elif fn.op in (ops.sum, ops.cum_sum):
            # postgres sometimes switches types for `sum`
            return sqa.cast(val, args[0].type)
        return val

    @classmethod
    def compile_ordered_aggregation(
        cls, *args: sqa.ColumnElement, order_by: list[sqa.UnaryExpression], impl
    ) -> sqa.ColumnElement:
        return impl(*args[:-1], aggregate_order_by(args[-1], *order_by))


with PostgresImpl.impl_store.impl_manager as impl:

    @impl(ops.less_than, String(), String())
    def _lt(x, y):
        return x < sqa.collate(y, "POSIX")

    @impl(ops.less_equal, String(), String())
    def _le(x, y):
        return x <= sqa.collate(y, "POSIX")

    @impl(ops.greater_than, String(), String())
    def _gt(x, y):
        return x > sqa.collate(y, "POSIX")

    @impl(ops.greater_equal, String(), String())
    def _ge(x, y):
        return x >= sqa.collate(y, "POSIX")

    @impl(ops.round)
    def _round(x, decimals=0):
        if decimals == 0:
            if isinstance(x.type, sqa.Integer):
                return x
            return sqa.func.ROUND(x, type_=x.type)

        if isinstance(x.type, sqa.Float):
            # Postgres doesn't support rounding of doubles to specific precision
            # -> Must first cast to numeric
            return sqa.func.ROUND(sqa.cast(x, sqa.Numeric), decimals, type_=sqa.Numeric).cast(x.type)

        return sqa.func.ROUND(x, decimals, type_=x.type)

    @impl(ops.dt_second)
    def _dt_second(x):
        return sqa.func.FLOOR(sqa.extract("second", x), type_=sqa.Integer())

    @impl(ops.dt_millisecond)
    def _dt_millisecond(x):
        return sqa.func.FLOOR(sqa.extract("millisecond", x), type_=sqa.Integer()) % 1000

    @impl(ops.dt_microsecond)
    def _dt_microsecond(x):
        return sqa.func.FLOOR(sqa.extract("microsecond", x), type_=sqa.Integer()) % 1_000_000

    @impl(ops.horizontal_max, String(), String(), ...)
    def _horizontal_max(*x):
        return sqa.func.GREATEST(*(sqa.collate(e, "POSIX") for e in x))

    @impl(ops.horizontal_min, String(), String(), ...)
    def _least(*x):
        return sqa.func.LEAST(*(sqa.collate(e, "POSIX") for e in x))

    @impl(ops.any)
    @impl(ops.max, Bool())
    def _any(x):
        return sqa.func.BOOL_OR(x, type_=sqa.Boolean())

    @impl(ops.all)
    @impl(ops.min, Bool())
    def _all(x):
        return sqa.func.BOOL_AND(x, type_=sqa.Boolean())

    @impl(ops.is_nan)
    def _is_nan(x):
        return x == PostgresImpl.nan()

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return x != PostgresImpl.nan()

    @impl(ops.dur_days)
    def _dur_days(x):
        sqa.func.extract("DAYS", x)

    @impl(ops.list_agg)
    def _list_agg(x):
        return sqa.func.array_agg(x)

    @impl(ops.truediv, Int(), Int())
    @impl(ops.truediv, Float(), Float())
    def _truediv(x, y):
        if not isinstance(x.type, Float):
            x = sqa.cast(x, sqa.Double)
        if not isinstance(y.type, Float):
            y = sqa.cast(y, sqa.Double)
        return x / y

    @impl(ops.dt_day_of_week)
    def _day_of_week(x):
        return (sqa.extract("dow", x) + 6) % sqa.literal_column("7") + 1

    @impl(ops.str_contains)
    def _str_contains(x, pattern, allow_regex, true_if_regex_unsupported):
        if not allow_regex:
            return x.contains(pattern, autoescape=True)
        return x.op("~")(pattern)
